# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ...tensor import arithmetic
from ...tensor.fuse import TensorCpFuseChunk
from ...utils import lazy_import
from .core import RuntimeOptimizer, register_optimizer


cp = lazy_import("cupy", rename="cp")
CP_INSTALLED = cp is not None

CP_ELEMENTWISE_OP = {
    arithmetic.TensorSubtract,
    arithmetic.TensorMultiply,
    arithmetic.TensorTrueDiv,
    arithmetic.TensorSqrt,
}
CP_OP = CP_ELEMENTWISE_OP


@register_optimizer
class CupyRuntimeOptimizer(RuntimeOptimizer):
    engine = "cupy"

    @classmethod
    def is_available(cls) -> bool:
        return CP_INSTALLED

    def optimize(self):
        fuses = []
        explored = set()

        graph = self._graph
        for node in graph.topological_iter():
            if type(node.op) not in CP_OP:
                continue
            if node in explored:
                continue
            if graph.count_predecessors(node) != 1:
                continue
            if node in graph.results:
                continue

            selected = [node]
            # add successors
            cur_node = graph.successors(node)[0]
            while (
                graph.count_predecessors(cur_node) == 1 and type(cur_node.op) in CP_OP
            ):
                selected.append(cur_node)
                if graph.count_successors(cur_node) != 1 or cur_node in graph.results:
                    break
                else:
                    cur_node = graph.successors(cur_node)[0]
            if len(selected) > 1:
                explored.update(selected)
                fuses.append(list(selected))

        return self._fuse_nodes(fuses, TensorCpFuseChunk)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import json
from typing import Any, BinaryIO, Dict, List, Optional, TextIO, Union

import numpy as np

from .... import opcodes as OperandDef
from ....core import recursive_tile
from ....core.context import get_context
from ....remote.run_script import RunScript, _extract_inputs
from ....serialization.serializables import (
    BytesField,
    Int32Field,
    DictField,
    StringField,
)
from ....typing import SessionType, TileableType
from ....utils import to_binary
from ...utils import collect_ports
from ..utils import pick_workers


class RunTensorFlow(RunScript):
    _op_type_ = OperandDef.RUN_TENSORFLOW

    _code = BytesField("code")
    _n_workers = Int32Field("n_workers")
    _n_ps = Int32Field("n_ps")
    _tf_config = DictField("tf_config")
    _port = Int32Field("port")
    # used for chunk op
    _tf_task_type = StringField("tf_task_type")
    _tf_task_index = Int32Field("tf_task_index")

    def __init__(
        self,
        n_workers=None,
        n_ps=None,
        tf_config=None,
        port=None,
        tf_task_type=None,
        tf_task_index=None,
        gpu=None,
        **kw,
    ):
        super().__init__(
            _n_workers=n_workers,
            _n_ps=n_ps,
            _tf_config=tf_config,
            _port=port,
            _tf_task_type=tf_task_type,
            _tf_task_index=tf_task_index,
            gpu=gpu,
            **kw,
        )

    @property
    def n_workers(self):
        return self._n_workers

    @property
    def n_ps(self):
        return self._n_ps or 0

    @property
    def n_roles(self):
        return self._n_workers + self._n_ps

    @property
    def tf_config(self):
        return self._tf_config

    @property
    def port(self):
        return self._port

    @property
    def tf_task_type(self):
        return self._tf_task_type

    @property
    def tf_task_index(self):
        return self._tf_task_index

    @classmethod
    def tile(cls, op):
        ctx = get_context()

        cluster_conf = {"worker": []}
        if op.n_ps > 0:
            cluster_conf["ps"] = []
        n_workers = op.n_workers

        out_chunks = []
        worker_addresses = ctx.get_worker_addresses()
        picked_workers = pick_workers(worker_addresses, op.n_roles)
        data, input_chunks = cls._get_chunk_data(op)

        ports = yield from recursive_tile(collect_ports(worker_addresses))
        yield ports.chunks
        ports = ctx.get_chunks_result([ports.chunks[0].key])[0]

        i = 0
        for worker, port in zip(picked_workers, ports):
            worker_addr = worker.rsplit(":", 1)[0]
            chunk_op = op.copy().reset_key()
            chunk_op._data = data
            addr = f"{worker_addr}:{port}"
            # tell graph actor that the chunk should be executed on the exact worker
            chunk_op.expect_worker = worker
            tp = "worker" if i < n_workers else "ps"
            chunk_op._tf_task_type = tp
            idx = i if i < n_workers else i - n_workers
            chunk_op._tf_task_index = idx
            cluster_conf[tp].append(addr)
            chunk_op._tf_config = {
                "cluster": cluster_conf,
                "task": {"type": tp, "index": idx},
            }
            out_chunks.append(chunk_op.new_chunk(input_chunks, index=(i,)))
            i += 1

        new_op = op.copy()
        return new_op.new_tileables(
            op.inputs,
            chunks=out_chunks,
            nsplits=(tuple(np.nan for _ in range(len(out_chunks))),),
        )

    @classmethod
    def _build_envs(cls, ctx, op):
        envs = super()._build_envs(ctx, op)
        envs.update({"TF_CONFIG": json.dumps(op.tf_config)})
        return envs

    @classmethod
    def execute(cls, ctx, op):
        if op.merge:
            return super().execute(ctx, op)

        assert ctx.local_address.split(":")[0] == op.expect_worker.split(":")[0]

        super().execute(ctx, op)

        if op.tf_task_type == "worker" and op.tf_task_index == 0:
            ctx[op.outputs[0].key] = {"status": "ok"}
        else:
            ctx[op.outputs[0].key] = {}


def run_tensorflow_script(
    script: Union[bytes, str, BinaryIO, TextIO],
    n_workers: int,
    n_ps: int = 0,
    data: Dict[str, TileableType] = None,
    gpu: Optional[bool] = None,
    command_argv: List[str] = None,
    retry_when_fail: bool = False,
    session: SessionType = None,
    run_kwargs: Dict[str, Any] = None,
):
    """
    Run TensorFlow script in Mars cluster.

    Parameters
    ----------
    script: str or file-like object
        Script to run
    n_workers : int
        Number of TensorFlow workers.
    n_ps : int
        Number of TensorFlow PS workers.
    data : dict
        Variable name to data.
    gpu : bool
        Run PyTorch script on GPU
    command_argv : list
        Extra command args for script
    retry_when_fail : bool
        If True, retry when function failed.
    session
        Mars session, if not provided, will use default one.
    run_kwargs : dict
        Extra kwargs for `session.run`.

    Returns
    -------
    status
        return {'status': 'ok'} if succeeded, or error raised
    """
    if int(n_workers) <= 0:
        raise ValueError("n_workers should be at least 1")
    if int(n_ps) < 0:
        raise ValueError("n_ps should be at least 0")
    if hasattr(script, "read"):
        code = script.read()
    else:
        with open(os.path.abspath(script), "rb") as f:
            code = f.read()

    inputs = _extract_inputs(data)
    op = RunTensorFlow(
        data=data,
        code=to_binary(code),
        n_workers=int(n_workers),
        n_ps=int(n_ps),
        retry_when_fail=retry_when_fail,
        gpu=gpu,
        command_args=command_argv,
    )
    return op(inputs).execute(session=session, **(run_kwargs or {}))



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import OrderedDict
from functools import reduce

import pandas as pd

from ..core import FuseChunkData, FuseChunk, ENTITY_TYPE, OutputType
from ..core.operand import (
    Operand,
    TileableOperandMixin,
    Fuse,
    ShuffleProxy,
    FuseChunkMixin,
)
from ..tensor.core import TENSOR_TYPE
from ..tensor.operands import TensorOperandMixin
from ..utils import calc_nsplits
from .core import (
    DATAFRAME_CHUNK_TYPE,
    SERIES_CHUNK_TYPE,
    INDEX_CHUNK_TYPE,
    DATAFRAME_TYPE,
    SERIES_TYPE,
    INDEX_TYPE,
    DATAFRAME_GROUPBY_TYPE,
    SERIES_GROUPBY_TYPE,
    CATEGORICAL_TYPE,
)
from .utils import parse_index


class DataFrameOperandMixin(TileableOperandMixin):
    __slots__ = ()
    _op_module_ = "dataframe"

    def new_dataframes(
        self,
        inputs,
        shape=None,
        dtypes=None,
        index_value=None,
        columns_value=None,
        chunks=None,
        nsplits=None,
        output_limit=None,
        kws=None,
        **kw
    ):
        setattr(self, "_output_types", [OutputType.dataframe])
        return self.new_tileables(
            inputs,
            shape=shape,
            dtypes=dtypes,
            index_value=index_value,
            columns_value=columns_value,
            chunks=chunks,
            nsplits=nsplits,
            output_limit=output_limit,
            kws=kws,
            **kw
        )

    def new_dataframe(
        self,
        inputs,
        shape=None,
        dtypes=None,
        index_value=None,
        columns_value=None,
        **kw
    ):
        if getattr(self, "output_limit") != 1:
            raise TypeError("cannot new DataFrame with more than 1 outputs")

        return self.new_dataframes(
            inputs,
            shape=shape,
            dtypes=dtypes,
            index_value=index_value,
            columns_value=columns_value,
            **kw
        )[0]

    def new_seriess(
        self,
        inputs,
        shape=None,
        dtype=None,
        index_value=None,
        name=None,
        chunks=None,
        nsplits=None,
        output_limit=None,
        kws=None,
        **kw
    ):
        setattr(self, "_output_types", [OutputType.series])
        return self.new_tileables(
            inputs,
            shape=shape,
            dtype=dtype,
            index_value=index_value,
            name=name,
            chunks=chunks,
            nsplits=nsplits,
            output_limit=output_limit,
            kws=kws,
            **kw
        )

    def new_series(
        self, inputs, shape=None, dtype=None, index_value=None, name=None, **kw
    ):
        if getattr(self, "output_limit") != 1:
            raise TypeError("cannot new Series with more than 1 outputs")

        return self.new_seriess(
            inputs, shape=shape, dtype=dtype, index_value=index_value, name=name, **kw
        )[0]

    def new_indexes(
        self,
        inputs,
        shape=None,
        dtype=None,
        index_value=None,
        name=None,
        chunks=None,
        nsplits=None,
        output_limit=None,
        kws=None,
        **kw
    ):
        setattr(self, "_output_types", [OutputType.index])
        return self.new_tileables(
            inputs,
            shape=shape,
            dtype=dtype,
            index_value=index_value,
            name=name,
            chunks=chunks,
            nsplits=nsplits,
            output_limit=output_limit,
            kws=kws,
            **kw
        )

    def new_index(
        self, inputs, shape=None, dtype=None, index_value=None, name=None, **kw
    ):
        if getattr(self, "output_limit") != 1:
            raise TypeError("cannot new Index with more than 1 outputs")

        return self.new_indexes(
            inputs, shape=shape, dtype=dtype, index_value=index_value, name=name, **kw
        )[0]

    def new_scalars(
        self, inputs, dtype=None, chunks=None, output_limit=None, kws=None, **kw
    ):
        setattr(self, "_output_types", [OutputType.scalar])
        return self.new_tileables(
            inputs,
            shape=(),
            dtype=dtype,
            chunks=chunks,
            nsplits=(),
            output_limit=output_limit,
            kws=kws,
            **kw
        )

    def new_scalar(self, inputs, dtype=None, **kw):
        if getattr(self, "output_limit") != 1:
            raise TypeError("cannot new tensor with more than 1 outputs")

        return self.new_scalars(inputs, dtype=dtype, **kw)[0]

    def new_categoricals(
        self,
        inputs,
        shape=None,
        dtype=None,
        categories_value=None,
        chunks=None,
        nsplits=None,
        output_limit=None,
        kws=None,
        **kw
    ):
        setattr(self, "_output_types", [OutputType.categorical])
        return self.new_tileables(
            inputs,
            shape=shape,
            dtype=dtype,
            categories_value=categories_value,
            chunks=chunks,
            nsplits=nsplits,
            output_limit=output_limit,
            kws=kws,
            **kw
        )

    def new_categorical(
        self, inputs, shape=None, dtype=None, categories_value=None, **kw
    ):
        if getattr(self, "output_limit") != 1:
            raise TypeError("cannot new Categorical with more than 1 outputs")

        return self.new_categoricals(
            inputs, shape=shape, dtype=dtype, categories_value=categories_value, **kw
        )[0]

    @classmethod
    def _process_groupby_params(cls, groupby_params):
        new_groupby_params = groupby_params.copy()
        if isinstance(groupby_params["by"], list):
            by = []
            for v in groupby_params["by"]:
                if isinstance(v, ENTITY_TYPE):
                    by.append(cls.concat_tileable_chunks(v).chunks[0])
                else:
                    by.append(v)
            new_groupby_params["by"] = by
        return new_groupby_params

    @classmethod
    def _get_groupby_inputs(cls, groupby, groupby_params):
        inputs = [groupby]
        chunk_inputs = list(groupby.chunks)
        if isinstance(groupby_params["by"], list):
            for chunk_v, v in zip(
                groupby_params["by"], groupby.op.groupby_params["by"]
            ):
                if isinstance(v, ENTITY_TYPE):
                    inputs.append(v)
                    chunk_inputs.append(chunk_v)
        return inputs, chunk_inputs

    @classmethod
    def concat_tileable_chunks(cls, tileable):
        from .merge.concat import DataFrameConcat, GroupByConcat

        df = tileable
        assert not df.is_coarse()

        if isinstance(df, DATAFRAME_TYPE):
            chunk = DataFrameConcat(output_types=[OutputType.dataframe]).new_chunk(
                df.chunks,
                shape=df.shape,
                index=(0, 0),
                dtypes=df.dtypes,
                index_value=df.index_value,
                columns_value=df.columns_value,
            )
            return DataFrameConcat(output_types=[OutputType.dataframe]).new_dataframe(
                [df],
                shape=df.shape,
                chunks=[chunk],
                nsplits=tuple((s,) for s in df.shape),
                dtypes=df.dtypes,
                index_value=df.index_value,
                columns_value=df.columns_value,
            )
        elif isinstance(df, SERIES_TYPE):
            chunk = DataFrameConcat(output_types=[OutputType.series]).new_chunk(
                df.chunks,
                shape=df.shape,
                index=(0,),
                dtype=df.dtype,
                index_value=df.index_value,
                name=df.name,
            )
            return DataFrameConcat(output_types=[OutputType.series]).new_series(
                [df],
                shape=df.shape,
                chunks=[chunk],
                nsplits=tuple((s,) for s in df.shape),
                dtype=df.dtype,
                index_value=df.index_value,
                name=df.name,
            )
        elif isinstance(df, INDEX_TYPE):
            chunk = DataFrameConcat(output_types=[OutputType.index]).new_chunk(
                df.chunks,
                shape=df.shape,
                index=(0,),
                dtype=df.dtype,
                index_value=df.index_value,
                name=df.name,
            )
            return DataFrameConcat(output_types=[OutputType.index]).new_index(
                [df],
                shape=df.shape,
                chunks=[chunk],
                nsplits=tuple((s,) for s in df.shape),
                dtype=df.dtype,
                index_value=df.index_value,
                name=df.name,
            )
        elif isinstance(df, (DATAFRAME_GROUPBY_TYPE, SERIES_GROUPBY_TYPE)):
            output_type = (
                OutputType.dataframe_groupby
                if isinstance(df, DATAFRAME_GROUPBY_TYPE)
                else OutputType.series_groupby
            )
            groupby_params = cls._process_groupby_params(df.op.groupby_params)
            inputs, chunk_inputs = cls._get_groupby_inputs(df, groupby_params)
            chunk = GroupByConcat(
                groups=df.chunks,
                groupby_params=groupby_params,
                output_types=[output_type],
            ).new_chunk(chunk_inputs, **df.params)
            return GroupByConcat(
                groups=[df],
                groupby_params=df.op.groupby_params,
                output_types=[output_type],
            ).new_tileable(inputs, chunks=[chunk], **df.params)
        elif isinstance(df, CATEGORICAL_TYPE):
            chunk = DataFrameConcat(output_types=[OutputType.categorical]).new_chunk(
                df.chunks,
                shape=df.shape,
                index=(0,),
                dtype=df.dtype,
                categories_value=df.categories_value,
            )
            return DataFrameConcat(
                output_types=[OutputType.categorical]
            ).new_categorical(
                [df],
                shape=df.shape,
                chunks=[chunk],
                nsplits=tuple((s,) for s in df.shape),
                dtype=df.dtype,
                categories_value=df.categories_value,
            )
        elif isinstance(df, TENSOR_TYPE):
            return TensorOperandMixin.concat_tileable_chunks(tileable)
        else:
            raise NotImplementedError

    @classmethod
    def create_tileable_from_chunks(cls, chunks, inputs=None, **kw):
        ndim = chunks[0].ndim
        index_min, index_max = [None] * ndim, [None] * ndim
        for c in chunks:
            for ax, i in enumerate(c.index):
                if index_min[ax] is None:
                    index_min[ax] = i
                else:
                    index_min[ax] = min(i, index_min[ax])
                if index_max[ax] is None:
                    index_max[ax] = i
                else:
                    index_max[ax] = max(i, index_max[ax])

        # gen {chunk index -> shape}
        chunk_index_to_shape = OrderedDict()
        chunk_index_to_chunk = dict()
        for c in chunks:
            new_index = []
            for ax, i in enumerate(c.index):
                new_index.append(i - index_min[ax])
            chunk_index_to_shape[tuple(new_index)] = c.shape
            chunk_index_to_chunk[tuple(new_index)] = c

        nsplits = calc_nsplits(chunk_index_to_shape)
        shape = tuple(sum(ns) for ns in nsplits)
        chunk_shape = tuple(len(ns) for ns in nsplits)
        op = chunks[0].op.copy().reset_key()
        if isinstance(chunks[0], DATAFRAME_CHUNK_TYPE):
            params = cls._calc_dataframe_params(chunk_index_to_chunk, chunk_shape)
            params.update(kw)
            return op.new_dataframe(
                inputs, shape=shape, chunks=chunks, nsplits=nsplits, **params
            )
        elif isinstance(chunks[0], SERIES_CHUNK_TYPE):
            params = cls._calc_series_index_params(chunks)
            params.update(kw)
            return op.new_series(
                inputs, shape=shape, chunks=chunks, nsplits=nsplits, **params
            )
        else:
            assert isinstance(chunks[0], INDEX_CHUNK_TYPE)
            params = cls._calc_series_index_params(chunks)
            params.update(kw)
            return op.new_index(
                inputs, shape=shape, chunks=chunks, nsplits=nsplits, **params
            )

    @classmethod
    def _calc_dataframe_params(cls, chunk_index_to_chunks, chunk_shape):
        dtypes = pd.concat(
            [
                chunk_index_to_chunks[0, i].dtypes
                for i in range(chunk_shape[1])
                if (0, i) in chunk_index_to_chunks
            ]
        )
        columns_value = parse_index(dtypes.index, store_data=True)
        pd_indexes = [
            chunk_index_to_chunks[i, 0].index_value.to_pandas()
            for i in range(chunk_shape[0])
            if (i, 0) in chunk_index_to_chunks
        ]
        pd_index = reduce(lambda x, y: x.append(y), pd_indexes)
        index_value = parse_index(pd_index)
        return {
            "dtypes": dtypes,
            "columns_value": columns_value,
            "index_value": index_value,
        }

    @classmethod
    def _calc_series_index_params(cls, chunks):
        pd_indexes = [c.index_value.to_pandas() for c in chunks]
        pd_index = reduce(lambda x, y: x.append(y), pd_indexes)
        index_value = parse_index(pd_index)
        return {"dtype": chunks[0].dtype, "index_value": index_value}

    def get_fuse_op_cls(self, _):
        return DataFrameFuseChunk


DataFrameOperand = Operand


class DataFrameShuffleProxy(ShuffleProxy, DataFrameOperandMixin):
    def __init__(self, sparse=None, output_types=None, **kwargs):
        super().__init__(sparse=sparse, _output_types=output_types, **kwargs)

    @classmethod
    def execute(cls, ctx, op):
        pass


class DataFrameFuseChunkMixin(FuseChunkMixin, DataFrameOperandMixin):
    __slots__ = ()

    def _create_chunk(self, output_idx, index, **kw):
        data = FuseChunkData(_index=index, _shape=kw.pop("shape", None), _op=self, **kw)

        return FuseChunk(data)


class DataFrameFuseChunk(Fuse, DataFrameFuseChunkMixin):
    @property
    def output_types(self):
        return self.outputs[-1].chunk.op.output_types



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np


def calc_svd_shapes(a):
    """
    Calculate output shapes of singular value decomposition.
    Follow the behavior of `numpy`:
    if a's shape is (6, 18), U's shape is (6, 6), s's shape is (6,), V's shape is (6, 18)
    if a's shape is (18, 6), U's shape is (18, 6), s's shape is (6,), V's shape is (6, 6)
    :param a: input tensor
    :return: (U.shape, s.shape, V.shape)
    """
    x, y = a.shape
    if x > y:
        return (x, y), (y,), (y, y)
    else:
        return (x, x), (x,), (x, y)


def svd_flip(u, v, u_based_decision=True):
    """
    Sign correction to ensure deterministic output from SVD.

    Adjusts the columns of u and the rows of v such that the loadings in the
    columns in u that are largest in absolute value are always positive.

    Parameters
    ----------
    u : Tensor
        u and v are the output of `linalg.svd` or
        `randomized_svd`, with matching inner dimensions
        so one can compute `mt.dot(u * s, v)`.

    v : Tensor
        u and v are the output of `linalg.svd` or
        `randomized_svd`, with matching inner dimensions
        so one can compute `mt.dot(u * s, v)`.

    u_based_decision : boolean, (default=True)
        If True, use the columns of u as the basis for sign flipping.
        Otherwise, use the rows of v. The choice of which variable to base the
        decision on is generally algorithm dependent.


    Returns
    -------
    u_adjusted, v_adjusted : arrays with the same dimensions as the input.

    """
    from ... import tensor as mt

    if u_based_decision:
        # columns of u, rows of v
        max_abs_cols = mt.argmax(mt.abs(u), axis=0)
        signs = mt.sign(u[max_abs_cols, np.arange(u.shape[1])])
        u *= signs
        v *= signs[:, mt.newaxis]
    else:
        # rows of v, columns of u
        max_abs_rows = mt.argmax(mt.abs(v), axis=1)
        signs = mt.sign(v[np.arange(v.shape[0]), max_abs_rows])
        u *= signs
        v *= signs[:, mt.newaxis]
    return u, v



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes as OperandDef
from ..datasource import tensor as astensor
from ..arithmetic.multiply import TensorTreeMultiply
from .core import TensorCumReduction, TensorCumReductionMixin


class TensorCumprod(TensorCumReduction, TensorCumReductionMixin):
    _op_type_ = OperandDef.CUMPROD
    _func_name = "cumprod"

    def __init__(self, axis=None, **kw):
        super().__init__(_axis=axis, **kw)

    @staticmethod
    def _get_op_types():
        return TensorCumprod, TensorTreeMultiply


def cumprod(a, axis=None, dtype=None, out=None):
    """
    Return the cumulative product of elements along a given axis.

    Parameters
    ----------
    a : array_like
        Input tensor.
    axis : int, optional
        Axis along which the cumulative product is computed.  By default
        the input is flattened.
    dtype : dtype, optional
        Type of the returned tensor, as well as of the accumulator in which
        the elements are multiplied.  If *dtype* is not specified, it
        defaults to the dtype of `a`, unless `a` has an integer dtype with
        a precision less than that of the default platform integer.  In
        that case, the default platform integer is used instead.
    out : Tensor, optional
        Alternative output tensor in which to place the result. It must
        have the same shape and buffer length as the expected output
        but the type of the resulting values will be cast if necessary.

    Returns
    -------
    cumprod : Tensor
        A new tensor holding the result is returned unless `out` is
        specified, in which case a reference to out is returned.

    See Also
    --------
    numpy.doc.ufuncs : Section "Output arguments"

    Notes
    -----
    Arithmetic is modular when using integer types, and no error is
    raised on overflow.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> a = mt.array([1,2,3])
    >>> mt.cumprod(a).execute() # intermediate results 1, 1*2
    ...                         # total product 1*2*3 = 6
    array([1, 2, 6])
    >>> a = mt.array([[1, 2, 3], [4, 5, 6]])
    >>> mt.cumprod(a, dtype=float).execute() # specify type of output
    array([   1.,    2.,    6.,   24.,  120.,  720.])

    The cumulative product for each column (i.e., over the rows) of `a`:

    >>> mt.cumprod(a, axis=0).execute()
    array([[ 1,  2,  3],
           [ 4, 10, 18]])

    The cumulative product for each row (i.e. over the columns) of `a`:

    >>> mt.cumprod(a,axis=1).execute()
    array([[  1,   2,   6],
           [  4,  20, 120]])

    """
    a = astensor(a)
    if dtype is None:
        dtype = np.empty((1,), dtype=a.dtype).cumprod().dtype
    op = TensorCumprod(axis=axis, dtype=dtype)
    return op(a, out=out)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, List, Tuple

import numpy as np

from .core import Serializer, buffered, pickle_buffers, unpickle_buffers


class NDArraySerializer(Serializer):
    @buffered
    def serial(self, obj: np.ndarray, context: Dict):
        header = {}
        if obj.dtype.hasobject:
            header["pickle"] = True
            buffers = pickle_buffers(obj)
            return (header,), buffers, True

        order = "C"
        if obj.flags.f_contiguous:
            order = "F"
        elif not obj.flags.c_contiguous:
            obj = np.ascontiguousarray(obj)
        try:
            desc = np.lib.format.dtype_to_descr(obj.dtype)
            dtype_new_order = None
        except ValueError:
            # for structured dtype, array[[field2, field1]] will create a view,
            # and dtype_to_desc will fail due to the order
            fields = obj.dtype.fields
            new_fields = sorted(fields, key=lambda k: fields[k][1])
            desc = np.lib.format.dtype_to_descr(obj.dtype[new_fields])
            dtype_new_order = list(fields)
        header.update(
            dict(
                pickle=False,
                descr=desc,
                dtype_new_order=dtype_new_order,
                shape=list(obj.shape),
                strides=list(obj.strides),
                order=order,
            )
        )
        return (header,), [memoryview(obj.ravel(order=order).view("uint8").data)], True

    def deserial(self, serialized: Tuple, context: Dict, subs: List[Any]):
        header = serialized[0]
        if header["pickle"]:
            return unpickle_buffers(subs)

        try:
            dtype = np.lib.format.descr_to_dtype(header["descr"])
        except AttributeError:  # pragma: no cover
            # for older numpy versions, descr_to_dtype is not implemented
            dtype = np.dtype(header["descr"])

        dtype_new_order = header["dtype_new_order"]
        if dtype_new_order:
            dtype = dtype[dtype_new_order]
        return np.ndarray(
            shape=tuple(header["shape"]),
            dtype=dtype,
            buffer=subs[0],
            strides=tuple(header["strides"]),
            order=header["order"],
        )


NDArraySerializer.register(np.ndarray)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
from collections import OrderedDict

from ....serialization.serializables import (
    Int64Field,
    BoolField,
    Int32Field,
    Float64Field,
    StringField,
)
from ....utils import pd_release_version
from ...utils import validate_axis
from ..core import Window

_default_min_period_1 = pd_release_version >= (1, 1, 0)
_pd_1_3_repr = pd_release_version >= (1, 3, 0)
_window_has_method = pd_release_version >= (1, 4, 0)


class EWM(Window):
    _alpha = Float64Field("alpha")
    _min_periods = Int64Field("min_periods")
    _adjust = BoolField("adjust")
    _ignore_na = BoolField("ignore_na")
    _axis = Int32Field("axis")
    _method = StringField("method")

    def __init__(
        self,
        alpha=None,
        min_periods=None,
        adjust=None,
        ignore_na=None,
        axis=None,
        method=None,
        **kw
    ):
        super().__init__(
            _alpha=alpha,
            _min_periods=min_periods,
            _adjust=adjust,
            _ignore_na=ignore_na,
            _axis=axis,
            _method=method or "single",
            **kw
        )

    @property
    def alpha(self):
        return self._alpha

    @property
    def min_periods(self):
        return self._min_periods

    @property
    def adjust(self):
        return self._adjust

    @property
    def ignore_na(self):
        return self._ignore_na

    @property
    def axis(self):
        return self._axis

    @property
    def method(self):
        return self._method

    @property
    def params(self):
        p = OrderedDict()

        if not _window_has_method:  # pragma: no cover
            args = ["alpha", "min_periods", "adjust", "ignore_na", "axis"]
        else:
            args = ["alpha", "min_periods", "adjust", "ignore_na", "axis", "method"]

        for k in args:
            p[k] = getattr(self, k)
        return p

    def __call__(self, df):
        return df.ewm(**self.params)

    def _repr(self, params):
        com = 1.0 / params.pop("alpha") - 1
        params["com"] = int(com) if _pd_1_3_repr and com == math.floor(com) else com
        try:
            params.move_to_end("com", last=False)
        except AttributeError:  # pragma: no cover
            pass
        return super()._repr(params)

    def _repr_name(self):
        try:
            from pandas.core.window import ExponentialMovingWindow  # noqa: F401

            return "ExponentialMovingWindow"
        except ImportError:  # pragma: no cover
            return "EWM"

    def aggregate(self, func):
        from .aggregation import DataFrameEwmAgg

        params = self.params
        params["alpha_ignore_na"] = params.pop("ignore_na", False)
        params["validate_columns"] = False
        op = DataFrameEwmAgg(func=func, **params)
        return op(self)

    agg = aggregate

    def mean(self):
        return self.aggregate("mean")

    def var(self):
        return self.aggregate("var")

    def std(self):
        return self.aggregate("std")


def ewm(
    obj,
    com=None,
    span=None,
    halflife=None,
    alpha=None,
    min_periods=0,
    adjust=True,
    ignore_na=False,
    axis=0,
):
    r"""
    Provide exponential weighted functions.

    Parameters
    ----------
    com : float, optional
        Specify decay in terms of center of mass,
        :math:`\alpha = 1 / (1 + com),\text{ for } com \geq 0`.
    span : float, optional
        Specify decay in terms of span,
        :math:`\alpha = 2 / (span + 1),\text{ for } span \geq 1`.
    halflife : float, optional
        Specify decay in terms of half-life,
        :math:`\alpha = 1 - exp(log(0.5) / halflife),\text{for} halflife > 0`.
    alpha : float, optional
        Specify smoothing factor :math:`\alpha` directly,
        :math:`0 < \alpha \leq 1`.
    min_periods : int, default 0
        Minimum number of observations in window required to have a value
        (otherwise result is NA).
    adjust : bool, default True
        Divide by decaying adjustment factor in beginning periods to account
        for imbalance in relative weightings
        (viewing EWMA as a moving average).
    ignore_na : bool, default False
        Ignore missing values when calculating weights;
        specify True to reproduce pre-0.15.0 behavior.
    axis : {0 or 'index', 1 or 'columns'}, default 0
        The axis to use. The value 0 identifies the rows, and 1
        identifies the columns.

    Returns
    -------
    DataFrame
        A Window sub-classed for the particular operation.

    See Also
    --------
    rolling : Provides rolling window calculations.
    expanding : Provides expanding transformations.

    Notes
    -----
    Exactly one of center of mass, span, half-life, and alpha must be provided.

    Allowed values and relationship between the parameters are specified in the
    parameter descriptions above; see the link at the end of this section for
    a detailed explanation.

    When adjust is True (default), weighted averages are calculated using
    weights (1-alpha)**(n-1), (1-alpha)**(n-2), ..., 1-alpha, 1.

    When adjust is False, weighted averages are calculated recursively as:

       weighted_average[0] = arg[0];
       weighted_average[i] = (1-alpha)*weighted_average[i-1] + alpha*arg[i].

    When ignore_na is False (default), weights are based on absolute positions.
    For example, the weights of x and y used in calculating the final weighted
    average of [x, None, y] are (1-alpha)**2 and 1 (if adjust is True), and
    (1-alpha)**2 and alpha (if adjust is False).

    When ignore_na is True (reproducing pre-0.15.0 behavior), weights are based
    on relative positions. For example, the weights of x and y used in
    calculating the final weighted average of [x, None, y] are 1-alpha and 1
    (if adjust is True), and 1-alpha and alpha (if adjust is False).

    More details can be found at
    https://pandas.pydata.org/pandas-docs/stable/user_guide/computation.html#exponentially-weighted-windows

    Examples
    --------
    >>> import numpy as np
    >>> import mars.dataframe as md
    >>> df = md.DataFrame({'B': [0, 1, 2, np.nan, 4]})
    >>> df.execute()
         B
    0  0.0
    1  1.0
    2  2.0
    3  NaN
    4  4.0
    >>> df.ewm(com=0.5).mean().execute()
              B
    0  0.000000
    1  0.750000
    2  1.615385
    3  1.615385
    4  3.670213
    """
    axis = validate_axis(axis, obj)

    decay_count = 0
    for arg in (com, span, halflife, alpha):
        if arg is not None:
            decay_count += 1

    if decay_count == 0:
        raise ValueError("Must pass one of comass, span, halflife, or alpha")
    if decay_count > 1:
        raise ValueError("comass, span, halflife, and alpha are mutually exclusive")

    if com is not None:
        if com < 0:
            raise ValueError("comass must satisfy: comass >= 0")
        alpha = 1.0 / (1 + com)
    elif span is not None:
        if span < 1:
            raise ValueError("span must satisfy: span >= 1")
        alpha = 2.0 / (1 + span)
    elif halflife is not None:
        if halflife <= 0:
            raise ValueError("halflife must satisfy: halflife > 0")
        alpha = 1.0 - math.exp(math.log(0.5) / halflife)
    if alpha <= 0 or alpha > 1:
        raise ValueError("alpha must satisfy: 0 < alpha <= 1")

    if not adjust and not ignore_na:
        raise NotImplementedError(
            "adjust == False when ignore_na == False not implemented"
        )
    if axis == 1:
        raise NotImplementedError("axis other than 0 is not supported")

    if alpha == 1:
        return obj.expanding(min_periods=min_periods, axis=axis)

    if _default_min_period_1:
        min_periods = min_periods or 1

    return EWM(
        input=obj,
        alpha=alpha,
        min_periods=min_periods,
        adjust=adjust,
        ignore_na=ignore_na,
        axis=axis,
    )



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

try:
    import scipy

    from .err_fresnel import (
        erf,
        TensorErf,
        erfc,
        TensorErfc,
        erfcx,
        TensorErfcx,
        erfi,
        TensorErfi,
        erfinv,
        TensorErfinv,
        erfcinv,
        TensorErfcinv,
    )
    from .gamma_funcs import (
        gamma,
        TensorGamma,
        gammaln,
        TensorGammaln,
        loggamma,
        TensorLogGamma,
        gammasgn,
        TensorGammaSgn,
        gammainc,
        TensorGammaInc,
        gammaincinv,
        TensorGammaIncInv,
        gammaincc,
        TensorGammaIncc,
        gammainccinv,
        TensorGammaInccInv,
        beta,
        TensorBeta,
        betaln,
        TensorBetaLn,
        betainc,
        TensorBetaInc,
        betaincinv,
        TensorBetaIncInv,
        psi,
        TensorPsi,
        rgamma,
        TensorRGamma,
        polygamma,
        TensorPolyGamma,
        multigammaln,
        TensorMultiGammaLn,
        digamma,
        TensorDiGamma,
        poch,
        TensorPoch,
    )
    from .info_theory import (
        entr,
        TensorEntr,
        rel_entr,
        TensorRelEntr,
        kl_div,
        TensorKlDiv,
    )
    from .convenience import (
        xlogy,
        TensorXLogY,
    )
    from .bessel import (
        jv,
        TensorJV,
        jve,
        TensorJVE,
        yn,
        TensorYN,
        yv,
        TensorYV,
        yve,
        TensorYVE,
        kn,
        TensorKN,
        kv,
        TensorKV,
        kve,
        TensorKVE,
        iv,
        TensorIV,
        ive,
        TensorIVE,
        hankel1,
        TensorHankel1,
        hankel1e,
        TensorHankel1e,
        hankel2,
        TensorHankel2,
        hankel2e,
        TensorHankel2e,
    )
    from .hypergeometric_funcs import (
        hyp2f1,
        TensorHYP2F1,
        hyp1f1,
        TensorHYP1F1,
        hyperu,
        TensorHYPERU,
        hyp0f1,
        TensorHYP0F1,
    )
    from .ellip_harm import (
        ellip_harm,
        TensorEllipHarm,
        ellip_harm_2,
        TensorEllipHarm2,
        ellip_normal,
        TensorEllipNormal,
    )
    from .ellip_func_integrals import (
        ellipk,
        TensorEllipk,
        ellipkm1,
        TensorEllipkm1,
        ellipkinc,
        TensorEllipkinc,
        ellipe,
        TensorEllipe,
        ellipeinc,
        TensorEllipeinc,
        elliprc,
        TensorElliprc,
        elliprd,
        TensorElliprd,
        elliprf,
        TensorElliprf,
        elliprg,
        TensorElliprg,
        elliprj,
        TensorElliprj,
    )
except ImportError:  # pragma: no cover
    pass

_names_to_del = [_name for _name, _val in globals().items() if _val is None]
[globals().pop(_name) for _name in _names_to_del]
del _names_to_del



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import Optional, Dict, Tuple

from ..metric import (
    AbstractMetric,
    AbstractCounter,
    AbstractGauge,
    AbstractHistogram,
    AbstractMeter,
)

logger = logging.getLogger(__name__)


class SimpleMetric:
    def __init__(
        self, name: str, description: str = "", tag_keys: Optional[Tuple[str]] = None
    ):
        self._name = name
        self._description = description
        self._tag_keys = tag_keys
        self._value = 0

    def update(self, value: float = 1.0, tags: Optional[Dict[str, str]] = None):
        self._value = value
        logger.debug(
            "Reporting metric with name: %s, description: %s, value: %s, tags: %s",
            self._name,
            self._description,
            value,
            tags,
        )

    @property
    def value(self):
        return self._value


class ConsoleMetricMixin(AbstractMetric):
    @property
    def value(self):
        return self._metric.value

    def _init(self):
        self._metric = SimpleMetric(self._name, self._description, self._tag_keys)

    def _record(self, value=1, tags: Optional[Dict[str, str]] = None):
        self._metric.update(value, tags)


class Counter(ConsoleMetricMixin, AbstractCounter):
    pass


class Gauge(ConsoleMetricMixin, AbstractGauge):
    pass


class Meter(ConsoleMetricMixin, AbstractMeter):
    pass


class Histogram(ConsoleMetricMixin, AbstractHistogram):
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..utils import validate_axis, check_out_param
from ..datasource import tensor as astensor


def take(a, indices, axis=None, out=None):
    """
    Take elements from a tensor along an axis.

    When axis is not None, this function does the same thing as "fancy"
    indexing (indexing arrays using tensors); however, it can be easier to use
    if you need elements along a given axis. A call such as
    ``mt.take(arr, indices, axis=3)`` is equivalent to
    ``arr[:,:,:,indices,...]``.

    Explained without fancy indexing, this is equivalent to the following use
    of `ndindex`, which sets each of ``ii``, ``jj``, and ``kk`` to a tuple of
    indices::

        Ni, Nk = a.shape[:axis], a.shape[axis+1:]
        Nj = indices.shape
        for ii in ndindex(Ni):
            for jj in ndindex(Nj):
                for kk in ndindex(Nk):
                    out[ii + jj + kk] = a[ii + (indices[jj],) + kk]

    Parameters
    ----------
    a : array_like (Ni..., M, Nk...)
        The source tensor.
    indices : array_like (Nj...)
        The indices of the values to extract.

        Also allow scalars for indices.
    axis : int, optional
        The axis over which to select values. By default, the flattened
        input tensor is used.
    out : Tensor, optional (Ni..., Nj..., Nk...)
        If provided, the result will be placed in this tensor. It should
        be of the appropriate shape and dtype.
    mode : {'raise', 'wrap', 'clip'}, optional
        Specifies how out-of-bounds indices will behave.

        * 'raise' -- raise an error (default)
        * 'wrap' -- wrap around
        * 'clip' -- clip to the range

        'clip' mode means that all indices that are too large are replaced
        by the index that addresses the last element along that axis. Note
        that this disables indexing with negative numbers.

    Returns
    -------
    out : Tensor (Ni..., Nj..., Nk...)
        The returned tensor has the same type as `a`.

    See Also
    --------
    compress : Take elements using a boolean mask
    Tensor.take : equivalent method

    Notes
    -----

    By eliminating the inner loop in the description above, and using `s_` to
    build simple slice objects, `take` can be expressed  in terms of applying
    fancy indexing to each 1-d slice::

        Ni, Nk = a.shape[:axis], a.shape[axis+1:]
        for ii in ndindex(Ni):
            for kk in ndindex(Nj):
                out[ii + s_[...,] + kk] = a[ii + s_[:,] + kk][indices]

    For this reason, it is equivalent to (but faster than) the following use
    of `apply_along_axis`::

        out = mt.apply_along_axis(lambda a_1d: a_1d[indices], axis, a)

    Examples
    --------
    >>> import mars.tensor as mt
    >>> a = [4, 3, 5, 7, 6, 8]
    >>> indices = [0, 1, 4]
    >>> mt.take(a, indices).execute()
    array([4, 3, 6])

    In this example if `a` is a tensor, "fancy" indexing can be used.

    >>> a = mt.array(a)
    >>> a[indices].execute()
    array([4, 3, 6])

    If `indices` is not one dimensional, the output also has these dimensions.

    >>> mt.take(a, [[0, 1], [2, 3]]).execute()
    array([[4, 3],
           [5, 7]])
    """
    a = astensor(a)
    if axis is None:
        t = a.ravel()[indices]
    else:
        axis = validate_axis(a.ndim, axis)
        t = a[(slice(None),) * axis + (indices,)]

    if out is None:
        return t

    if out.shape != t.shape:
        raise ValueError(
            f"output tensor has wrong shape, expect: {t.shape}, got: {out.shape}"
        )
    check_out_param(out, t, "unsafe")
    out.data = t.data
    return out



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect

import numpy as np
import pandas as pd

from ... import opcodes
from ...config import options
from ...core import OutputType, recursive_tile
from ...core.custom_log import redirect_custom_log
from ...core.operand import OperatorLogicKeyGeneratorMixin
from ...serialization.serializables import (
    StringField,
    AnyField,
    BoolField,
    TupleField,
    DictField,
    FunctionField,
)
from ...utils import enter_current_session, quiet_stdio, get_func_token_values
from ..arrays import ArrowArray
from ..operands import DataFrameOperandMixin, DataFrameOperand
from ..utils import (
    build_df,
    build_series,
    parse_index,
    validate_axis,
    validate_output_types,
    make_dtypes,
    make_dtype,
    build_empty_df,
    build_empty_series,
)


class ApplyOperandLogicKeyGeneratorMixin(OperatorLogicKeyGeneratorMixin):
    def _get_logic_key_token_values(self):
        token_values = super()._get_logic_key_token_values() + [
            self._axis,
            self._convert_dtype,
            self._raw,
            self._result_type,
            self._elementwise,
        ]
        if self.func:
            return token_values + get_func_token_values(self.func)
        else:  # pragma: no cover
            return token_values


class ApplyOperand(
    DataFrameOperand, DataFrameOperandMixin, ApplyOperandLogicKeyGeneratorMixin
):
    _op_type_ = opcodes.APPLY

    _func = FunctionField("func")
    _axis = AnyField("axis")
    _convert_dtype = BoolField("convert_dtype")
    _raw = BoolField("raw")
    _result_type = StringField("result_type")
    _elementwise = BoolField("elementwise")
    _args = TupleField("args")
    _kwds = DictField("kwds")

    def __init__(
        self,
        func=None,
        axis=None,
        convert_dtype=None,
        raw=None,
        result_type=None,
        args=None,
        kwds=None,
        output_type=None,
        elementwise=None,
        **kw,
    ):
        if output_type:
            kw["_output_types"] = [output_type]
        super().__init__(
            _func=func,
            _axis=axis,
            _convert_dtype=convert_dtype,
            _raw=raw,
            _result_type=result_type,
            _args=args,
            _kwds=kwds,
            _elementwise=elementwise,
            **kw,
        )

    @property
    def func(self):
        return self._func

    @property
    def axis(self):
        return self._axis

    @property
    def convert_dtype(self):
        return self._convert_dtype

    @property
    def raw(self):
        return self._raw

    @property
    def result_type(self):
        return self._result_type

    @property
    def elementwise(self):
        return self._elementwise

    @property
    def args(self):
        return getattr(self, "_args", None) or ()

    @property
    def kwds(self):
        return getattr(self, "_kwds", None) or dict()

    @classmethod
    @redirect_custom_log
    @enter_current_session
    def execute(cls, ctx, op):
        input_data = ctx[op.inputs[0].key]
        out = op.outputs[0]
        if len(input_data) == 0:
            if op.output_types[0] == OutputType.dataframe:
                ctx[out.key] = build_empty_df(out.dtypes)
            else:
                ctx[out.key] = build_empty_series(out.dtype, name=out.name)
            return

        if isinstance(input_data, pd.DataFrame):
            result = input_data.apply(
                op.func,
                axis=op.axis,
                raw=op.raw,
                result_type=op.result_type,
                args=op.args,
                **op.kwds,
            )
        else:
            try:
                result = input_data.apply(
                    op.func, convert_dtype=op.convert_dtype, args=op.args, **op.kwds
                )
            except TypeError:
                if isinstance(input_data.values, ArrowArray):
                    input_data = pd.Series(
                        input_data.to_numpy(),
                        name=input_data.name,
                        index=input_data.index,
                    )
                    result = input_data.apply(
                        op.func, convert_dtype=op.convert_dtype, args=op.args, **op.kwds
                    )
                else:  # pragma: no cover
                    raise
        ctx[out.key] = result

    @classmethod
    def _tile_df(cls, op):
        in_df = op.inputs[0]
        out_df = op.outputs[0]
        axis = op.axis
        elementwise = op.elementwise

        if not elementwise and in_df.chunk_shape[axis] > 1:
            chunk_size = (
                in_df.shape[axis],
                max(1, options.chunk_store_limit // in_df.shape[axis]),
            )
            if axis == 1:
                chunk_size = chunk_size[::-1]
            in_df = yield from recursive_tile(in_df.rechunk(chunk_size))

        chunks = []
        if out_df.ndim == 2:
            for c in in_df.chunks:
                if elementwise:
                    new_shape = c.shape
                    new_index_value, new_columns_value = c.index_value, c.columns_value
                else:
                    new_shape = [np.nan, np.nan]
                    new_shape[1 - axis] = c.shape[1 - axis]
                    if axis == 0:
                        new_index_value = out_df.index_value
                        new_columns_value = c.columns_value
                    else:
                        new_index_value = c.index_value
                        new_columns_value = out_df.columns_value

                if op.axis == 0:
                    new_dtypes = out_df.dtypes[c.dtypes.keys()]
                else:
                    new_dtypes = out_df.dtypes

                new_op = op.copy().reset_key()
                new_op.tileable_op_key = op.key
                chunks.append(
                    new_op.new_chunk(
                        [c],
                        shape=tuple(new_shape),
                        index=c.index,
                        dtypes=new_dtypes,
                        index_value=new_index_value,
                        columns_value=new_columns_value,
                    )
                )

            new_nsplits = list(in_df.nsplits)
            if not elementwise:
                new_nsplits[axis] = (np.nan,) * len(new_nsplits[axis])
        else:
            for c in in_df.chunks:
                shape_len = c.shape[1 - axis]
                new_index_value = c.index_value if axis == 1 else c.columns_value
                new_index = (c.index[1 - axis],)
                new_op = op.copy().reset_key()
                new_op.tileable_op_key = op.key
                chunks.append(
                    new_op.new_chunk(
                        [c],
                        shape=(shape_len,),
                        index=new_index,
                        dtype=out_df.dtype,
                        index_value=new_index_value,
                    )
                )
            new_nsplits = (in_df.nsplits[1 - axis],)

        new_op = op.copy()
        kw = out_df.params.copy()
        kw.update(dict(chunks=chunks, nsplits=tuple(new_nsplits)))
        return new_op.new_tileables(op.inputs, **kw)

    @classmethod
    def _tile_series(cls, op):
        in_series = op.inputs[0]
        out_series = op.outputs[0]

        chunks = []
        for c in in_series.chunks:
            new_op = op.copy().reset_key()
            new_op.tileable_op_key = op.key
            kw = c.params.copy()
            if out_series.ndim == 1:
                kw["dtype"] = out_series.dtype
            else:
                kw["index"] = (c.index[0], 0)
                kw["shape"] = (c.shape[0], out_series.shape[1])
                kw["dtypes"] = out_series.dtypes
                kw["columns_value"] = out_series.columns_value
            chunks.append(new_op.new_chunk([c], **kw))

        new_op = op.copy()
        kw = out_series.params.copy()
        kw.update(dict(chunks=chunks, nsplits=in_series.nsplits))
        if out_series.ndim == 2:
            kw["nsplits"] = (in_series.nsplits[0], (out_series.shape[1],))
            kw["columns_value"] = out_series.columns_value
        return new_op.new_tileables(op.inputs, **kw)

    @classmethod
    def tile(cls, op):
        if op.inputs[0].ndim == 2:
            return (yield from cls._tile_df(op))
        else:
            return cls._tile_series(op)

    def _infer_df_func_returns(self, df, dtypes, dtype=None, name=None, index=None):
        if isinstance(self._func, np.ufunc):
            output_type = OutputType.dataframe
            new_dtypes = None
            index_value = "inherit"
            new_elementwise = True
        else:
            if self.output_types is not None and (
                dtypes is not None or dtype is not None
            ):
                ret_dtypes = dtypes if dtypes is not None else (name, dtype)
                ret_index_value = parse_index(index) if index is not None else None
                self._elementwise = False
                return ret_dtypes, ret_index_value

            output_type = new_dtypes = index_value = None
            new_elementwise = False

        try:
            empty_df = build_df(df, size=2)
            with np.errstate(all="ignore"), quiet_stdio():
                infer_df = empty_df.apply(
                    self._func,
                    axis=self._axis,
                    raw=self._raw,
                    result_type=self._result_type,
                    args=self.args,
                    **self.kwds,
                )
            if index_value is None:
                if infer_df.index is empty_df.index:
                    index_value = "inherit"
                else:
                    index_value = parse_index(pd.RangeIndex(-1))

            if isinstance(infer_df, pd.DataFrame):
                output_type = output_type or OutputType.dataframe
                new_dtypes = new_dtypes or infer_df.dtypes
            else:
                output_type = output_type or OutputType.series
                new_dtypes = (name or infer_df.name, dtype or infer_df.dtype)
            new_elementwise = False if new_elementwise is None else new_elementwise
        except:  # noqa: E722  # nosec
            pass

        self.output_types = (
            [output_type] if not self.output_types else self.output_types
        )
        dtypes = new_dtypes if dtypes is None else dtypes
        index_value = index_value if index is None else parse_index(index)
        self._elementwise = (
            new_elementwise if self._elementwise is None else self._elementwise
        )
        return dtypes, index_value

    def _call_dataframe(self, df, dtypes=None, dtype=None, name=None, index=None):
        # for backward compatibility
        dtype = dtype if dtype is not None else dtypes
        dtypes, index_value = self._infer_df_func_returns(
            df, dtypes, dtype=dtype, name=name, index=index
        )
        if index_value is None:
            index_value = parse_index(None, (df.key, df.index_value.key))
        for arg, desc in zip((self.output_types, dtypes), ("output_types", "dtypes")):
            if arg is None:
                raise TypeError(
                    f"Cannot determine {desc} by calculating with enumerate data, "
                    "please specify it as arguments"
                )

        if index_value == "inherit":
            index_value = df.index_value

        if self._elementwise:
            shape = df.shape
        elif self.output_types[0] == OutputType.dataframe:
            shape = [np.nan, np.nan]
            shape[1 - self.axis] = df.shape[1 - self.axis]
            shape = tuple(shape)
        else:
            shape = (df.shape[1 - self.axis],)

        if self.output_types[0] == OutputType.dataframe:
            if self.axis == 0:
                return self.new_dataframe(
                    [df],
                    shape=shape,
                    dtypes=dtypes,
                    index_value=index_value,
                    columns_value=parse_index(dtypes.index, store_data=True),
                )
            else:
                return self.new_dataframe(
                    [df],
                    shape=shape,
                    dtypes=dtypes,
                    index_value=df.index_value,
                    columns_value=parse_index(dtypes.index, store_data=True),
                )
        else:
            name, dtype = dtypes
            return self.new_series(
                [df], shape=shape, name=name, dtype=dtype, index_value=index_value
            )

    def _call_series(self, series, dtypes=None, dtype=None, name=None, index=None):
        # for backward compatibility
        dtype = dtype if dtype is not None else dtypes
        if self._convert_dtype:
            if self.output_types is not None and (
                dtypes is not None or dtype is not None
            ):
                infer_series = test_series = None
            else:
                test_series = build_series(series, size=2, name=series.name)
                try:
                    with np.errstate(all="ignore"), quiet_stdio():
                        infer_series = test_series.apply(
                            self._func, args=self.args, **self.kwds
                        )
                except:  # noqa: E722  # nosec  # pylint: disable=bare-except
                    infer_series = None

            output_type = self._output_types[0]

            if index is not None:
                index_value = parse_index(index)
            elif infer_series is not None:
                if infer_series.index is test_series.index:
                    index_value = series.index_value
                else:  # pragma: no cover
                    index_value = parse_index(infer_series.index)
            else:
                index_value = parse_index(None, series)

            if output_type == OutputType.dataframe:
                if dtypes is None:
                    if infer_series is not None and infer_series.ndim == 2:
                        dtypes = infer_series.dtypes
                    else:
                        raise TypeError(
                            "Cannot determine dtypes, "
                            "please specify `dtypes` as argument"
                        )
                columns_value = parse_index(dtypes.index, store_data=True)

                return self.new_dataframe(
                    [series],
                    shape=(series.shape[0], len(dtypes)),
                    index_value=index_value,
                    columns_value=columns_value,
                    dtypes=dtypes,
                )
            else:
                if (
                    dtype is None
                    and infer_series is not None
                    and infer_series.ndim == 1
                ):
                    dtype = infer_series.dtype
                else:
                    dtype = dtype if dtype is not None else np.dtype(object)
                if infer_series is not None and infer_series.ndim == 1:
                    name = name or infer_series.name
                return self.new_series(
                    [series],
                    dtype=dtype,
                    shape=series.shape,
                    index_value=index_value,
                    name=name,
                )
        else:
            dtype = dtype if dtype is not None else np.dtype("object")
            return self.new_series(
                [series],
                dtype=dtype,
                shape=series.shape,
                index_value=series.index_value,
                name=name,
            )

    def __call__(self, df_or_series, dtypes=None, dtype=None, name=None, index=None):
        axis = getattr(self, "axis", None) or 0
        dtypes = make_dtypes(dtypes)
        dtype = make_dtype(dtype)
        self._axis = validate_axis(axis, df_or_series)

        if df_or_series.op.output_types[0] == OutputType.dataframe:
            return self._call_dataframe(
                df_or_series, dtypes=dtypes, dtype=dtype, name=name, index=index
            )
        else:
            return self._call_series(
                df_or_series, dtypes=dtypes, dtype=dtype, name=name, index=index
            )


def df_apply(
    df,
    func,
    axis=0,
    raw=False,
    result_type=None,
    args=(),
    dtypes=None,
    dtype=None,
    name=None,
    output_type=None,
    index=None,
    elementwise=None,
    **kwds,
):
    """
    Apply a function along an axis of the DataFrame.

    Objects passed to the function are Series objects whose index is
    either the DataFrame's index (``axis=0``) or the DataFrame's columns
    (``axis=1``). By default (``result_type=None``), the final return type
    is inferred from the return type of the applied function. Otherwise,
    it depends on the `result_type` argument.

    Parameters
    ----------
    func : function
        Function to apply to each column or row.
    axis : {0 or 'index', 1 or 'columns'}, default 0
        Axis along which the function is applied:

        * 0 or 'index': apply function to each column.
        * 1 or 'columns': apply function to each row.

    raw : bool, default False
        Determines if row or column is passed as a Series or ndarray object:

        * ``False`` : passes each row or column as a Series to the
          function.
        * ``True`` : the passed function will receive ndarray objects
          instead.
          If you are just applying a NumPy reduction function this will
          achieve much better performance.

    result_type : {'expand', 'reduce', 'broadcast', None}, default None
        These only act when ``axis=1`` (columns):

        * 'expand' : list-like results will be turned into columns.
        * 'reduce' : returns a Series if possible rather than expanding
          list-like results. This is the opposite of 'expand'.
        * 'broadcast' : results will be broadcast to the original shape
          of the DataFrame, the original index and columns will be
          retained.

        The default behaviour (None) depends on the return value of the
        applied function: list-like results will be returned as a Series
        of those. However if the apply function returns a Series these
        are expanded to columns.

    output_type : {'dataframe', 'series'}, default None
        Specify type of returned object. See `Notes` for more details.

    dtypes : Series, default None
        Specify dtypes of returned DataFrames. See `Notes` for more details.

    dtype : numpy.dtype, default None
        Specify dtype of returned Series. See `Notes` for more details.

    name : str, default None
        Specify name of returned Series. See `Notes` for more details.

    index : Index, default None
        Specify index of returned object. See `Notes` for more details.

    elementwise : bool, default False
        Specify whether ``func`` is an elementwise function:

        * ``False`` : The function is not elementwise. Mars will try
          concatenating chunks in rows (when ``axis=0``) or in columns
          (when ``axis=1``) and then apply ``func`` onto the concatenated
          chunk. The concatenation step can cause extra latency.
        * ``True`` : The function is elementwise. Mars will apply
          ``func`` to original chunks. This will not introduce extra
          concatenation step and reduce overhead.

    args : tuple
        Positional arguments to pass to `func` in addition to the
        array/series.

    **kwds
        Additional keyword arguments to pass as keywords arguments to
        `func`.

    Returns
    -------
    Series or DataFrame
        Result of applying ``func`` along the given axis of the
        DataFrame.

    See Also
    --------
    DataFrame.applymap: For elementwise operations.
    DataFrame.aggregate: Only perform aggregating type operations.
    DataFrame.transform: Only perform transforming type operations.

    Notes
    -----
    When deciding output dtypes and shape of the return value, Mars will
    try applying ``func`` onto a mock DataFrame,  and the apply call may
    fail. When this happens, you need to specify the type of apply call
    (DataFrame or Series) in output_type.

    * For DataFrame output, you need to specify a list or a pandas Series
      as ``dtypes`` of output DataFrame. ``index`` of output can also be
      specified.
    * For Series output, you need to specify ``dtype`` and ``name`` of
      output Series.

    Examples
    --------
    >>> import numpy as np
    >>> import mars.tensor as mt
    >>> import mars.dataframe as md
    >>> df = md.DataFrame([[4, 9]] * 3, columns=['A', 'B'])
    >>> df.execute()
       A  B
    0  4  9
    1  4  9
    2  4  9

    Using a reducing function on either axis

    >>> df.apply(np.sum, axis=0).execute()
    A    12
    B    27
    dtype: int64

    >>> df.apply(np.sum, axis=1).execute()
    0    13
    1    13
    2    13
    dtype: int64

    Returning a list-like will result in a Series

    >>> df.apply(lambda x: [1, 2], axis=1).execute()
    0    [1, 2]
    1    [1, 2]
    2    [1, 2]
    dtype: object

    Passing ``result_type='expand'`` will expand list-like results
    to columns of a Dataframe

    >>> df.apply(lambda x: [1, 2], axis=1, result_type='expand').execute()
       0  1
    0  1  2
    1  1  2
    2  1  2

    Returning a Series inside the function is similar to passing
    ``result_type='expand'``. The resulting column names
    will be the Series index.

    >>> df.apply(lambda x: md.Series([1, 2], index=['foo', 'bar']), axis=1).execute()
       foo  bar
    0    1    2
    1    1    2
    2    1    2

    Passing ``result_type='broadcast'`` will ensure the same shape
    result, whether list-like or scalar is returned by the function,
    and broadcast it along the axis. The resulting column names will
    be the originals.

    >>> df.apply(lambda x: [1, 2], axis=1, result_type='broadcast').execute()
       A  B
    0  1  2
    1  1  2
    2  1  2
    """
    if isinstance(func, (list, dict)):
        return df.aggregate(func)

    output_types = kwds.pop("output_types", None)
    object_type = kwds.pop("object_type", None)
    output_types = validate_output_types(
        output_type=output_type, output_types=output_types, object_type=object_type
    )
    output_type = output_types[0] if output_types else None

    # calling member function
    if isinstance(func, str):
        func = getattr(df, func)
        sig = inspect.getfullargspec(func)
        if "axis" in sig.args:
            kwds["axis"] = axis
        return func(*args, **kwds)

    op = ApplyOperand(
        func=func,
        axis=axis,
        raw=raw,
        result_type=result_type,
        args=args,
        kwds=kwds,
        output_type=output_type,
        elementwise=elementwise,
    )
    return op(df, dtypes=dtypes, dtype=dtype, name=name, index=index)


def series_apply(
    series,
    func,
    convert_dtype=True,
    output_type=None,
    args=(),
    dtypes=None,
    dtype=None,
    name=None,
    index=None,
    **kwds,
):
    """
    Invoke function on values of Series.

    Can be ufunc (a NumPy function that applies to the entire Series)
    or a Python function that only works on single values.

    Parameters
    ----------
    func : function
        Python function or NumPy ufunc to apply.

    convert_dtype : bool, default True
        Try to find better dtype for elementwise function results. If
        False, leave as dtype=object.

    output_type : {'dataframe', 'series'}, default None
        Specify type of returned object. See `Notes` for more details.

    dtypes : Series, default None
        Specify dtypes of returned DataFrames. See `Notes` for more details.

    dtype : numpy.dtype, default None
        Specify dtype of returned Series. See `Notes` for more details.

    name : str, default None
        Specify name of returned Series. See `Notes` for more details.

    index : Index, default None
        Specify index of returned object. See `Notes` for more details.

    args : tuple
        Positional arguments passed to func after the series value.

    **kwds
        Additional keyword arguments passed to func.

    Returns
    -------
    Series or DataFrame
        If func returns a Series object the result will be a DataFrame.

    See Also
    --------
    Series.map: For element-wise operations.
    Series.agg: Only perform aggregating type operations.
    Series.transform: Only perform transforming type operations.

    Notes
    -----
    When deciding output dtypes and shape of the return value, Mars will
    try applying ``func`` onto a mock Series, and the apply call may fail.
    When this happens, you need to specify the type of apply call
    (DataFrame or Series) in output_type.

    * For DataFrame output, you need to specify a list or a pandas Series
      as ``dtypes`` of output DataFrame. ``index`` of output can also be
      specified.
    * For Series output, you need to specify ``dtype`` and ``name`` of
      output Series.

    Examples
    --------
    Create a series with typical summer temperatures for each city.

    >>> import mars.tensor as mt
    >>> import mars.dataframe as md
    >>> s = md.Series([20, 21, 12],
    ...               index=['London', 'New York', 'Helsinki'])
    >>> s.execute()
    London      20
    New York    21
    Helsinki    12
    dtype: int64

    Square the values by defining a function and passing it as an
    argument to ``apply()``.

    >>> def square(x):
    ...     return x ** 2
    >>> s.apply(square).execute()
    London      400
    New York    441
    Helsinki    144
    dtype: int64

    Square the values by passing an anonymous function as an
    argument to ``apply()``.

    >>> s.apply(lambda x: x ** 2).execute()
    London      400
    New York    441
    Helsinki    144
    dtype: int64

    Define a custom function that needs additional positional
    arguments and pass these additional arguments using the
    ``args`` keyword.

    >>> def subtract_custom_value(x, custom_value):
    ...     return x - custom_value

    >>> s.apply(subtract_custom_value, args=(5,)).execute()
    London      15
    New York    16
    Helsinki     7
    dtype: int64

    Define a custom function that takes keyword arguments
    and pass these arguments to ``apply``.

    >>> def add_custom_values(x, **kwargs):
    ...     for month in kwargs:
    ...         x += kwargs[month]
    ...     return x

    >>> s.apply(add_custom_values, june=30, july=20, august=25).execute()
    London      95
    New York    96
    Helsinki    87
    dtype: int64
    """
    if isinstance(func, (list, dict)):
        return series.aggregate(func)

    # calling member function
    if isinstance(func, str):
        func_body = getattr(series, func, None)
        if func_body is not None:
            return func_body(*args, **kwds)
        func_str = func
        func = getattr(np, func_str, None)
        if func is None:
            raise AttributeError(
                f"'{func_str!r}' is not a valid function "
                f"for '{type(series).__name__}' object"
            )

    output_types = kwds.pop("output_types", None)
    object_type = kwds.pop("object_type", None)
    output_types = validate_output_types(
        output_type=output_type, output_types=output_types, object_type=object_type
    )
    output_type = output_types[0] if output_types else OutputType.series

    op = ApplyOperand(
        func=func,
        convert_dtype=convert_dtype,
        args=args,
        kwds=kwds,
        output_type=output_type,
    )
    return op(series, dtypes=dtypes, dtype=dtype, name=name, index=index)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Union

from .... import oscar as mo
from ....core import Tileable
from ....lib.aio import alru_cache
from ...subtask import SubtaskResult
from ..core import TileableGraph, TaskResult
from ..supervisor.manager import TaskManagerActor
from .core import AbstractTaskAPI


class TaskAPI(AbstractTaskAPI):
    def __init__(
        self, session_id: str, task_manager_ref: mo.ActorRefType[TaskManagerActor]
    ):
        self._session_id = session_id
        self._task_manager_ref = task_manager_ref

    @classmethod
    @alru_cache(cache_exceptions=False)
    async def create(cls, session_id: str, address: str) -> "TaskAPI":
        """
        Create Task API.

        Parameters
        ----------
        session_id : str
            Session ID
        address : str
            Supervisor address.

        Returns
        -------
        task_api
            Task API.
        """
        task_manager_ref = await mo.actor_ref(
            address, TaskManagerActor.gen_uid(session_id)
        )
        return TaskAPI(session_id, task_manager_ref)

    async def get_task_results(self, progress: bool = False) -> List[TaskResult]:
        return await self._task_manager_ref.get_task_results(progress)

    async def submit_tileable_graph(
        self,
        graph: TileableGraph,
        task_name: str = None,
        fuse_enabled: bool = None,
        extra_config: dict = None,
    ) -> str:
        try:
            return await self._task_manager_ref.submit_tileable_graph(
                graph, task_name, fuse_enabled=fuse_enabled, extra_config=extra_config
            )
        except mo.ActorNotExist:
            raise RuntimeError("Session closed already")

    async def get_tileable_graph_as_json(self, task_id: str):
        return await self._task_manager_ref.get_tileable_graph_dict_by_task_id(task_id)

    async def get_tileable_details(self, task_id: str):
        return await self._task_manager_ref.get_tileable_details(task_id)

    async def get_tileable_subtasks(
        self, task_id: str, tileable_id: str, with_input_output: bool
    ):

        return await self._task_manager_ref.get_tileable_subtasks(
            task_id, tileable_id, with_input_output
        )

    async def wait_task(self, task_id: str, timeout: float = None):
        return await self._task_manager_ref.wait_task(task_id, timeout=timeout)

    async def get_task_result(self, task_id: str) -> TaskResult:
        return await self._task_manager_ref.get_task_result(task_id)

    async def get_task_progress(self, task_id: str) -> float:
        return await self._task_manager_ref.get_task_progress(task_id)

    async def cancel_task(self, task_id: str):
        return await self._task_manager_ref.cancel_task(task_id)

    async def get_fetch_tileables(self, task_id: str) -> List[Tileable]:
        return await self._task_manager_ref.get_task_result_tileables(task_id)

    async def set_subtask_result(self, subtask_result: SubtaskResult):
        return await self._task_manager_ref.set_subtask_result.tell(subtask_result)

    async def get_last_idle_time(self) -> Union[float, None]:
        return await self._task_manager_ref.get_last_idle_time()



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
from typing import Any, Dict, List, Tuple, Union, Type

import numpy as np
import pandas as pd

from ...typing import BandType
from ...utils import dataslots, TypeDispatcher

PandasDtypeType = Union[np.dtype, pd.api.extensions.ExtensionDtype]

_meta_class_dispatcher = TypeDispatcher()


def register_meta_type(object_types: Tuple):
    def _call(meta_type: Type["_CommonMeta"]):
        _meta_class_dispatcher.register(object_types, meta_type)
        return meta_type

    return _call


def get_meta_type(object_type: Type) -> Type["_CommonMeta"]:
    return _meta_class_dispatcher.get_handler(object_type)


@dataslots
@dataclass
class _CommonMeta:
    """
    Class for common meta, for both tileable and chunk, or DataFrame, tensor etc.
    """

    object_id: str
    name: Any = None
    memory_size: int = None  # size in memory
    store_size: int = None  # size that stored in storage
    extra: Dict = None

    def merge_from(self, value: "_CommonMeta"):
        return self


@dataslots
@dataclass
class _TileableMeta(_CommonMeta):
    nsplits: Tuple[Tuple[int]] = None


@dataslots
@dataclass
class _ChunkMeta(_CommonMeta):
    index: Tuple[int] = None
    bands: List[BandType] = None
    # needed by ray ownership to keep object alive when worker died.
    object_refs: List[Any] = None

    def merge_from(self, value: "_ChunkMeta"):
        if value.bands:
            self.bands = list(set(self.bands) | set(value.bands))
        if value.object_refs:
            self.object_refs = list(set(self.object_refs) | set(value.object_refs))
        return self



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from numpy.linalg import LinAlgError

from ... import opcodes as OperandDef
from ...core import recursive_tile
from ...serialization.serializables import KeyField
from ..array_utils import as_same_device, device
from ..datasource import tensor as astensor
from ..operands import TensorHasInput, TensorOperandMixin
from ..core import TensorOrder


class TensorInv(TensorHasInput, TensorOperandMixin):
    _op_type_ = OperandDef.INV

    _input = KeyField("input")

    def __call__(self, a):
        a = astensor(a)
        return self.new_tensor([a], a.shape, order=TensorOrder.C_ORDER)

    @classmethod
    def _tile_one_chunk(cls, op):
        out = op.outputs[0]
        chunk_op = op.copy().reset_key()
        chunk_params = out.params
        chunk_params["index"] = (0,) * out.ndim
        out_chunk = chunk_op.new_chunk(op.inputs[0].chunks, kws=[chunk_params])

        new_op = op.copy()
        params = out.params
        params["nsplits"] = tuple((s,) for s in out.shape)
        params["chunks"] = [out_chunk]
        return new_op.new_tensors(op.inputs, kws=[params])

    @classmethod
    def tile(cls, op):
        """
        Use LU decomposition to compute inverse of matrix.
        Given a square matrix A:
        P, L, U = lu(A)
        b_eye is an identity matrix with the same shape as matrix A, then,
        (P * L * U) * A_inv = b_eye
        L * (U * A_inv) = P.T * b_eye
        use `solve_triangular` twice to compute the inverse of matrix A.
        """
        from .lu import lu
        from ..datasource import eye
        from ..base.transpose import TensorTranspose
        from .tensordot import tensordot
        from .solve_triangular import solve_triangular

        in_tensor = op.input
        is_sparse = in_tensor.is_sparse()

        if len(in_tensor.chunks) == 1:
            return cls._tile_one_chunk(op)

        b_eye = eye(in_tensor.shape[0], chunk_size=in_tensor.nsplits, sparse=is_sparse)

        p, l, u = lu(in_tensor)

        # transposed p equals to inverse of p
        p_transpose = TensorTranspose(
            dtype=p.dtype, sparse=p.op.sparse, axes=list(range(in_tensor.ndim))[::-1]
        ).new_tensor([p], p.shape)

        b = tensordot(
            p_transpose, b_eye, axes=((p_transpose.ndim - 1,), (b_eye.ndim - 2,))
        )

        # as `l` is a lower matrix, `lower=True` should be specified.
        uy = solve_triangular(l, b, lower=True, sparse=op.sparse)

        a_inv = solve_triangular(u, uy, sparse=op.sparse)
        a_inv = yield from recursive_tile(a_inv)
        return [a_inv]

    @classmethod
    def execute(cls, ctx, op):
        (inp,), device_id, xp = as_same_device(
            [ctx[c.key] for c in op.inputs], device=op.device, ret_extra=True
        )

        with device(device_id):
            ctx[op.outputs[0].key] = xp.linalg.inv(inp)


def inv(a, sparse=None):
    """
    Compute the (multiplicative) inverse of a matrix.
    Given a square matrix `a`, return the matrix `ainv` satisfying
    ``dot(a, ainv) = dot(ainv, a) = eye(a.shape[0])``.

    Parameters
    ----------
    a : (..., M, M) array_like
        Matrix to be inverted.
    sparse: bool, optional
        Return sparse value or not.

    Returns
    -------
    ainv : (..., M, M) ndarray or matrix
        (Multiplicative) inverse of the matrix `a`.

    Raises
    ------
    LinAlgError
        If `a` is not square or inversion fails.

    Examples
    --------
    >>> import mars.tensor as mt
    >>> a = np.array([[1., 2.], [3., 4.]])
    >>> ainv = mt.linalg.inv(a)
    >>> mt.allclose(mt.dot(a, ainv), mt.eye(2)).execute()
    True

    >>> mt.allclose(mt.dot(ainv, a), mt.eye(2)).execute()
    True

    >>> ainv.execute()
    array([[ -2. ,  1. ],
           [ 1.5, -0.5]])
    """

    # TODO: using some parallel algorithm for matrix inversion.
    a = astensor(a)
    if a.ndim != 2:
        raise LinAlgError(
            f"{a.ndim}-dimensional array given. Tensor must be two-dimensional"
        )
    if a.shape[0] != a.shape[1]:
        raise LinAlgError("Input must be square")

    tiny_inv = np.linalg.inv(np.array([[1, 2], [2, 5]], dtype=a.dtype))
    sparse = sparse if sparse is not None else a.issparse()
    op = TensorInv(dtype=tiny_inv.dtype, sparse=sparse)
    return op(a)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from typing import List, Union

from ....core import Tileable
from ..core import TileableGraph, TaskResult


class AbstractTaskAPI(ABC):
    @abstractmethod
    async def get_task_results(self, progress: bool = False) -> List[TaskResult]:
        """
        Get results of all tasks in the session

        Parameters
        ----------
        progress : bool
            If True, will return task progress

        Returns
        -------
        task_results: List[TaskResult]
            List of task results
        """

    @abstractmethod
    async def submit_tileable_graph(
        self,
        graph: TileableGraph,
        task_name: str = None,
        fuse_enabled: bool = True,
        extra_config: dict = None,
    ) -> str:
        """
        Submit a tileable graph

        Parameters
        ----------
        graph : TileableGraph
            Tileable graph.
        task_name : str
            Task name
        fuse_enabled : bool
            Enable fuse optimization
        extra_config : dict
            Extra config.

        Returns
        -------
        task_id : str
            Task ID.
        """

    @abstractmethod
    async def wait_task(self, task_id: str, timeout: float = None):
        """
        Wait for a task to finish.

        Parameters
        ----------
        task_id : str
            Task ID
        timeout: float
            Second to timeout
        """

    @abstractmethod
    async def cancel_task(self, task_id: str):
        """
        Cancel task.

        Parameters
        ----------
        task_id : str
            Task ID.
        """

    @abstractmethod
    async def get_task_result(self, task_id: str) -> TaskResult:
        """
        Get task status.

        Parameters
        ----------
        task_id : str
            Task ID.

        Returns
        -------
        result : TaskResult
            Task result.
        """

    @abstractmethod
    async def get_task_progress(self, task_id: str) -> float:
        """
        Get task progress.

        Parameters
        ----------
        task_id : str
            Task ID.

        Returns
        -------
        progress : float
            Get task progress.
        """

    @abstractmethod
    async def get_fetch_tileables(self, task_id: str) -> List[Tileable]:
        """
        Get fetch tileable for a task.

        Parameters
        ----------
        task_id : str
            Task ID.

        Returns
        -------
        fetch_tileable_list
            Fetch tileable list.
        """

    @abstractmethod
    async def get_last_idle_time(self) -> Union[float, None]:
        """
        Get last idle time from task manager.

        Returns
        -------
        last_idle_time: float
            The last idle time if the task manager is idle else None.
        """



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import functools
import logging
import operator
import sys
from dataclasses import dataclass
from typing import List, Dict, Any, Set, Callable
from .....core import ChunkGraph, Chunk, TileContext
from .....core.context import set_context
from .....core.operand import (
    Fetch,
    Fuse,
    VirtualOperand,
    MapReduceOperand,
    execute,
)
from .....lib.aio import alru_cache
from .....lib.ordered_set import OrderedSet
from .....resource import Resource
from .....serialization import serialize, deserialize
from .....typing import BandType
from .....utils import (
    calc_data_size,
    lazy_import,
    get_chunk_params,
    get_chunk_key_to_data_keys,
    ensure_coverage,
)
from ....lifecycle.api import LifecycleAPI
from ....meta.api import MetaAPI
from ....subtask import Subtask, SubtaskGraph
from ....subtask.utils import iter_input_data_keys, iter_output_data
from ...core import Task
from ..api import (
    TaskExecutor,
    ExecutionChunkResult,
    register_executor_cls,
)
from .config import RayExecutionConfig, IN_RAY_CI
from .context import (
    RayExecutionContext,
    RayExecutionWorkerContext,
    RayRemoteObjectManager,
)

ray = lazy_import("ray")
logger = logging.getLogger(__name__)


@dataclass
class _RayChunkMeta:
    memory_size: int


class RayTaskState(RayRemoteObjectManager):
    @classmethod
    def _gen_name(cls, task_id: str):
        return f"{cls.__name__}_{task_id}"

    @classmethod
    def get_handle(cls, task_id: str):
        """Get the RayTaskState actor handle."""
        name = cls._gen_name(task_id)
        logger.info("Getting %s handle.", name)
        return ray.get_actor(name)

    @classmethod
    def create(cls, task_id: str):
        """Create a RayTaskState actor."""
        name = cls._gen_name(task_id)
        logger.info("Creating %s.", name)
        return ray.remote(cls).options(name=name).remote()


_optimize_physical = None


def _optimize_subtask_graph(subtask_graph):
    global _optimize_physical

    if _optimize_physical is None:
        from .....optimization.physical import optimize as _optimize_physical
    return _optimize_physical(subtask_graph)


async def _cancel_ray_task(obj_ref, kill_timeout: int = 3):
    await asyncio.to_thread(ray.cancel, obj_ref, force=False)
    try:
        await asyncio.to_thread(ray.get, obj_ref, timeout=kill_timeout)
    except ray.exceptions.TaskCancelledError:  # pragma: no cover
        logger.info("Cancel ray task %s successfully.", obj_ref)
    except BaseException as e:
        logger.info(
            "Failed to cancel ray task %s with exception %s, "
            "force cancel the task by killing the worker.",
            e,
            obj_ref,
        )
        await asyncio.to_thread(ray.cancel, obj_ref, force=True)


def execute_subtask(
    task_id: str,
    subtask_id: str,
    subtask_chunk_graph: ChunkGraph,
    output_meta_keys: Set[str],
    input_keys: List[str],
    *inputs,
):
    logger.info("Begin to execute subtask: %s", subtask_id)
    ensure_coverage()
    subtask_chunk_graph = deserialize(*subtask_chunk_graph)
    # inputs = [i[1] for i in inputs]
    context = RayExecutionWorkerContext(
        lambda: RayTaskState.get_handle(task_id), zip(input_keys, inputs)
    )
    # optimize chunk graph.
    subtask_chunk_graph = _optimize_subtask_graph(subtask_chunk_graph)
    # from data_key to results
    for chunk in subtask_chunk_graph.topological_iter():
        if chunk.key not in context:
            execute(context, chunk.op)

    output = {
        key: data for key, data, _ in iter_output_data(subtask_chunk_graph, context)
    }
    output_values = []
    if output_meta_keys:
        output_meta = {}
        for chunk in subtask_chunk_graph.result_chunks:
            chunk_key = chunk.key
            if chunk_key in output_meta_keys and chunk_key not in output_meta:
                if isinstance(chunk.op, Fuse):
                    # fuse op
                    chunk = chunk.chunk
                data = context[chunk_key]
                memory_size = calc_data_size(data)
                output_meta[chunk_key] = get_chunk_params(chunk), memory_size
        assert len(output_meta_keys) == len(output_meta)
        output_values.append(output_meta)
    output_values.extend(output.values())

    logger.info("Finish executing subtask: %s", subtask_id)
    return output_values[0] if len(output_values) == 1 else output_values


@register_executor_cls
class RayTaskExecutor(TaskExecutor):
    name = "ray"

    def __init__(
        self,
        config: RayExecutionConfig,
        task: Task,
        tile_context: TileContext,
        task_context: Dict[str, "ray.ObjectRef"],
        task_chunks_meta: Dict[str, _RayChunkMeta],
        lifecycle_api: LifecycleAPI,
        meta_api: MetaAPI,
    ):
        self._config = config
        self._task = task
        self._tile_context = tile_context
        self._task_context = task_context
        self._task_chunks_meta = task_chunks_meta
        self._ray_executor = self._get_ray_executor()

        # api
        self._lifecycle_api = lifecycle_api
        self._meta_api = meta_api

        self._available_band_resources = None

        # For progress and task cancel
        self._pre_all_stages_progress = 0.0
        self._pre_all_stages_tile_progress = 0.0
        self._cur_stage_progress = 0.0
        self._cur_stage_tile_progress = 0.0
        self._cur_stage_first_output_object_ref_to_subtask = dict()
        self._execute_subtask_graph_aiotask = None
        self._cancelled = False

    @classmethod
    async def create(
        cls,
        config: RayExecutionConfig,
        *,
        session_id: str,
        address: str,
        task: Task,
        tile_context: TileContext,
        **kwargs,
    ) -> "RayTaskExecutor":
        lifecycle_api, meta_api = await cls._get_apis(session_id, address)
        task_context = {}
        task_chunks_meta = {}

        executor = cls(
            config,
            task,
            tile_context,
            task_context,
            task_chunks_meta,
            lifecycle_api,
            meta_api,
        )
        available_band_resources = await executor.get_available_band_resources()
        worker_addresses = list(
            map(operator.itemgetter(0), available_band_resources.keys())
        )
        if config.create_task_state_actor_as_needed():
            create_task_state_actor = lambda: RayTaskState.create(  # noqa: E731
                task_id=task.task_id
            )
        else:
            actor_handle = RayTaskState.create(task_id=task.task_id)
            create_task_state_actor = lambda: actor_handle  # noqa: E731
        await cls._init_context(
            config,
            task_context,
            task_chunks_meta,
            create_task_state_actor,
            worker_addresses,
            session_id,
            address,
        )
        return executor

    # noinspection DuplicatedCode
    def destroy(self):
        self._config = None
        self._task = None
        self._tile_context = None
        self._task_context = {}
        self._task_chunks_meta = {}
        self._ray_executor = None

        # api
        self._lifecycle_api = None
        self._meta_api = None

        self._available_band_resources = None

        # For progress and task cancel
        self._pre_all_stages_progress = 1.0
        self._pre_all_stages_tile_progress = 1.0
        self._cur_stage_progress = 1.0
        self._cur_stage_tile_progress = 1.0
        self._cur_stage_first_output_object_ref_to_subtask = dict()
        self._execute_subtask_graph_aiotask = None
        self._cancelled = None

    @classmethod
    @alru_cache(cache_exceptions=False)
    async def _get_apis(cls, session_id: str, address: str):
        return await asyncio.gather(
            LifecycleAPI.create(session_id, address),
            MetaAPI.create(session_id, address),
        )

    @staticmethod
    @functools.lru_cache(maxsize=None)  # Specify maxsize=None to make it faster
    def _get_ray_executor():
        # Export remote function once.
        return ray.remote(execute_subtask)

    @classmethod
    async def _init_context(
        cls,
        config: RayExecutionConfig,
        task_context: Dict[str, "ray.ObjectRef"],
        task_chunks_meta: Dict[str, _RayChunkMeta],
        create_task_state_actor: Callable[[], "ray.actor.ActorHandle"],
        worker_addresses: List[str],
        session_id: str,
        address: str,
    ):
        loop = asyncio.get_running_loop()
        context = RayExecutionContext(
            config,
            task_context,
            task_chunks_meta,
            worker_addresses,
            create_task_state_actor,
            session_id,
            address,
            address,
            address,
            loop=loop,
        )
        await context.init()
        set_context(context)

    async def execute_subtask_graph(
        self,
        stage_id: str,
        subtask_graph: SubtaskGraph,
        chunk_graph: ChunkGraph,
        tile_context: TileContext,
        context: Any = None,
    ) -> Dict[Chunk, ExecutionChunkResult]:
        if self._cancelled is True:  # pragma: no cover
            raise asyncio.CancelledError()
        logger.info("Stage %s start.", stage_id)
        # Make sure each stage use a clean dict.
        self._cur_stage_first_output_object_ref_to_subtask = dict()

        def _on_monitor_aiotask_done(fut):
            # Print the error of monitor task.
            try:
                fut.result()
            except asyncio.CancelledError:
                pass
            except Exception:  # pragma: no cover
                logger.exception(
                    "The monitor task of stage %s is done with exception.", stage_id
                )
                if IN_RAY_CI:  # pragma: no cover
                    logger.warning(
                        "The process will be exit due to the monitor task exception "
                        "when MARS_CI_BACKEND=ray."
                    )
                    sys.exit(-1)

        result_meta_keys = {
            chunk.key
            for chunk in chunk_graph.result_chunks
            if not isinstance(chunk.op, Fetch)
        }
        # Create a monitor task to update progress and collect garbage.
        monitor_aiotask = asyncio.create_task(
            self._update_progress_and_collect_garbage(
                stage_id,
                subtask_graph,
                result_meta_keys,
                self._config.get_subtask_monitor_interval(),
            )
        )
        monitor_aiotask.add_done_callback(_on_monitor_aiotask_done)

        def _on_execute_aiotask_done(_):
            # Make sure the monitor task is cancelled.
            monitor_aiotask.cancel()
            # Just use `self._cur_stage_tile_progress` as current stage progress
            # because current stage is completed, its progress is 1.0.
            self._cur_stage_progress = 1.0
            self._pre_all_stages_progress += self._cur_stage_tile_progress

        self._execute_subtask_graph_aiotask = asyncio.current_task()
        self._execute_subtask_graph_aiotask.add_done_callback(_on_execute_aiotask_done)

        task_context = self._task_context
        output_meta_object_refs = []
        self._pre_all_stages_tile_progress = (
            self._pre_all_stages_tile_progress + self._cur_stage_tile_progress
        )
        self._cur_stage_tile_progress = (
            self._tile_context.get_all_progress() - self._pre_all_stages_tile_progress
        )
        logger.info("Submitting %s subtasks of stage %s.", len(subtask_graph), stage_id)
        subtask_max_retries = self._config.get_subtask_max_retries()
        for subtask in subtask_graph.topological_iter():
            subtask_chunk_graph = subtask.chunk_graph
            key_to_input = await self._load_subtask_inputs(
                stage_id, subtask, subtask_chunk_graph, task_context
            )
            output_keys = self._get_subtask_output_keys(subtask_chunk_graph)
            output_meta_keys = result_meta_keys & output_keys
            output_count = len(output_keys) + bool(output_meta_keys)
            max_retries = subtask_max_retries if subtask.retryable else 0
            output_object_refs = self._ray_executor.options(
                num_returns=output_count, max_retries=max_retries
            ).remote(
                subtask.task_id,
                subtask.subtask_id,
                serialize(subtask_chunk_graph),
                output_meta_keys,
                list(key_to_input.keys()),
                *key_to_input.values(),
            )
            if output_count == 0:
                continue
            elif output_count == 1:
                output_object_refs = [output_object_refs]
            self._cur_stage_first_output_object_ref_to_subtask[
                output_object_refs[0]
            ] = subtask
            if output_meta_keys:
                meta_object_ref, *output_object_refs = output_object_refs
                # TODO(fyrestone): Fetch(not get) meta object here.
                output_meta_object_refs.append(meta_object_ref)
            task_context.update(zip(output_keys, output_object_refs))
        logger.info("Submitted %s subtasks of stage %s.", len(subtask_graph), stage_id)

        key_to_meta = {}
        if len(output_meta_object_refs) > 0:
            # TODO(fyrestone): Optimize update meta by fetching partial meta.
            meta_count = len(output_meta_object_refs)
            logger.info("Getting %s metas of stage %s.", meta_count, stage_id)
            meta_list = await asyncio.gather(*output_meta_object_refs)
            for meta in meta_list:
                for key, (params, memory_size) in meta.items():
                    key_to_meta[key] = params
                    self._task_chunks_meta[key] = _RayChunkMeta(memory_size=memory_size)
            assert len(key_to_meta) == len(result_meta_keys)
            logger.info("Got %s metas of stage %s.", meta_count, stage_id)

        chunk_to_meta = {}
        # ray.wait requires the object ref list is unique.
        output_object_refs = set()
        for chunk in chunk_graph.result_chunks:
            chunk_key = chunk.key
            object_ref = task_context[chunk_key]
            output_object_refs.add(object_ref)
            chunk_params = key_to_meta.get(chunk_key)
            if chunk_params is not None:
                chunk_to_meta[chunk] = ExecutionChunkResult(chunk_params, object_ref)

        logger.info("Waiting for stage %s complete.", stage_id)
        # Patched the asyncio.to_thread for Python < 3.9 at mars/lib/aio/__init__.py
        await asyncio.to_thread(ray.wait, list(output_object_refs), fetch_local=False)

        logger.info("Stage %s is complete.", stage_id)
        return chunk_to_meta

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if exc_type is not None:
            try:
                await self.cancel()
            except BaseException:  # noqa: E722  # nosec  # pylint: disable=bare-except
                pass
            return

        # Update info if no exception occurs.
        tileable_keys = []
        update_metas = []
        update_lifecycles = []
        for tileable in self._task.tileable_graph.result_tileables:
            tileable_keys.append(tileable.key)
            tileable = tileable.data if hasattr(tileable, "data") else tileable
            chunk_keys = []
            for chunk in self._tile_context[tileable].chunks:
                chunk_key = chunk.key
                chunk_keys.append(chunk_key)
                if chunk_key in self._task_context:
                    # Some tileable graph may have result chunks that not be executed,
                    # for example:
                    # r, b = cut(series, bins, retbins=True)
                    #     r_result = r.execute().fetch()
                    #     b_result = b.execute().fetch() <- This is the case
                    object_ref = self._task_context[chunk_key]
                    chunk_meta = self._task_chunks_meta[chunk_key]
                    update_metas.append(
                        self._meta_api.set_chunk_meta.delay(
                            chunk,
                            bands=[],
                            object_ref=object_ref,
                            memory_size=chunk_meta.memory_size,
                        )
                    )
                update_lifecycles.append(
                    self._lifecycle_api.track.delay(tileable.key, chunk_keys)
                )
        await self._meta_api.set_chunk_meta.batch(*update_metas)
        await self._lifecycle_api.track.batch(*update_lifecycles)
        await self._lifecycle_api.incref_tileables(tileable_keys)

    async def get_available_band_resources(self) -> Dict[BandType, Resource]:
        if self._available_band_resources is None:
            band_resources = self._config.get_band_resources()
            virtual_band_resources = {}
            idx = 0
            for band_resource in band_resources:
                for band, resource in band_resource.items():
                    virtual_band_resources[
                        (f"ray_virtual_address_{idx}:0", band)
                    ] = resource
                    idx += 1
            self._available_band_resources = virtual_band_resources

        return self._available_band_resources

    async def get_progress(self) -> float:
        """Get the execution progress."""
        return self._cur_stage_progress

    async def cancel(self):
        """
        Cancel the task execution.

        1. Try to cancel the `execute_subtask_graph`
        2. Try to cancel the submitted subtasks by `ray.cancel`
        """
        logger.info("Start to cancel task %s.", self._task)
        if self._task is None or self._cancelled is True:
            return
        self._cancelled = True
        if self._execute_subtask_graph_aiotask is not None:
            self._execute_subtask_graph_aiotask.cancel()
        timeout = self._config.get_subtask_cancel_timeout()
        to_be_cancelled_coros = [
            _cancel_ray_task(object_ref, timeout)
            for object_ref in self._cur_stage_first_output_object_ref_to_subtask.keys()
        ]
        await asyncio.gather(*to_be_cancelled_coros)

    async def _load_subtask_inputs(
        self, stage_id: str, subtask: Subtask, chunk_graph: ChunkGraph, context: Dict
    ):
        """
        Load a dict of input key to object ref of subtask from context.

        It updates the context if the input object refs are fetched from
        the meta service.
        """
        key_to_input = {}
        key_to_get_meta = {}
        chunk_key_to_data_keys = get_chunk_key_to_data_keys(chunk_graph)
        for key, _ in iter_input_data_keys(
            subtask, chunk_graph, chunk_key_to_data_keys
        ):
            if key in context:
                key_to_input[key] = context[key]
            else:
                key_to_get_meta[key] = self._meta_api.get_chunk_meta.delay(
                    key, fields=["object_refs"]
                )
        if key_to_get_meta:
            logger.info(
                "Fetch %s metas and update context of stage %s.",
                len(key_to_get_meta),
                stage_id,
            )
            meta_list = await self._meta_api.get_chunk_meta.batch(
                *key_to_get_meta.values()
            )
            for key, meta in zip(key_to_get_meta.keys(), meta_list):
                object_ref = meta["object_refs"][0]
                key_to_input[key] = object_ref
                context[key] = object_ref
        return key_to_input

    @staticmethod
    def _get_subtask_output_keys(chunk_graph: ChunkGraph):
        output_keys = {}
        for chunk in chunk_graph.results:
            if isinstance(chunk.op, VirtualOperand):
                continue
            elif isinstance(chunk.op, MapReduceOperand):
                # TODO(fyrestone): Handle shuffle operands.
                raise NotImplementedError(
                    "The shuffle operands are not supported by the ray executor."
                )
            else:
                output_keys[chunk.key] = 1
        return output_keys.keys()

    async def _update_progress_and_collect_garbage(
        self,
        stage_id: str,
        subtask_graph: SubtaskGraph,
        result_meta_keys: Set[str],
        interval_seconds: float,
    ):
        object_ref_to_subtask = self._cur_stage_first_output_object_ref_to_subtask
        total = len(subtask_graph)
        completed_subtasks = OrderedSet()

        def gc():
            """
            Consume the completed subtasks and collect garbage.

            GC the output object refs of the subtask which successors are submitted
            (not completed as above) can reduce the memory peaks, but we can't cancel
            and rerun slow subtasks because the input object refs of running subtasks
            may be deleted.
            """
            i = 0
            gc_subtasks = set()

            while i < total:
                while i >= len(completed_subtasks):
                    yield
                # Iterate the completed subtasks once.
                subtask = completed_subtasks[i]
                i += 1
                logger.debug("GC[stage=%s]: %s", stage_id, subtask)

                # Note: There may be a scenario in which delayed gc occurs.
                # When a subtask has more than one predecessor, like A, B,
                # and in the `for ... in ...` loop we get A firstly while
                # B's successors are completed, A's not. Then we cannot remove
                # B's results chunks before A's.
                for pred in subtask_graph.iter_predecessors(subtask):
                    if pred in gc_subtasks:
                        continue
                    while not all(
                        succ in completed_subtasks
                        for succ in subtask_graph.iter_successors(pred)
                    ):
                        yield
                    for chunk in pred.chunk_graph.results:
                        chunk_key = chunk.key
                        # We need to check the GC chunk key is not in the
                        # result meta keys, because there are some special
                        # cases that the result meta keys are not the leaves.
                        #
                        # example: test_cut_execution
                        if chunk_key not in result_meta_keys:
                            logger.debug("GC[stage=%s]: %s", stage_id, chunk)
                            self._task_context.pop(chunk_key, None)
                    gc_subtasks.add(pred)

            # TODO(fyrestone): Check the remaining self._task_context.keys()
            # in the result subtasks

        collect_garbage = gc()

        while len(completed_subtasks) < total:
            if len(object_ref_to_subtask) <= 0:  # pragma: no cover
                await asyncio.sleep(interval_seconds)

            # Only wait for unready subtask object refs.
            ready_objects, _ = await asyncio.to_thread(
                ray.wait,
                list(object_ref_to_subtask.keys()),
                num_returns=len(object_ref_to_subtask),
                timeout=0,
                fetch_local=False,
            )
            if len(ready_objects) == 0:
                await asyncio.sleep(interval_seconds)
                continue

            # Pop the completed subtasks from object_ref_to_subtask.
            completed_subtasks.update(map(object_ref_to_subtask.pop, ready_objects))
            # Update progress.
            stage_progress = (
                len(completed_subtasks) / total * self._cur_stage_tile_progress
            )
            self._cur_stage_progress = self._pre_all_stages_progress + stage_progress
            # Collect garbage, use `for ... in ...` to avoid raising StopIteration.
            for _ in collect_garbage:
                break
            # Fast to next loop and give it a chance to update object_ref_to_subtask.
            await asyncio.sleep(0)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes as OperandDef
from ..utils import infer_dtype
from .core import TensorUnaryOp
from .utils import arithmetic_operand


@arithmetic_operand(sparse_mode="unary")
class TensorArccos(TensorUnaryOp):
    _op_type_ = OperandDef.ARCCOS

    _func_name = "arccos"


@infer_dtype(np.arccos)
def arccos(x, out=None, where=None, **kwargs):
    """
    Trigonometric inverse cosine, element-wise.

    The inverse of `cos` so that, if ``y = cos(x)``, then ``x = arccos(y)``.

    Parameters
    ----------
    x : array_like
        `x`-coordinate on the unit circle.
        For real arguments, the domain is [-1, 1].
    out : Tensor, None, or tuple of Tensor and None, optional
        A location into which the result is stored. If provided, it must have
        a shape that the inputs broadcast to. If not provided or `None`,
        a freshly-allocated tensor is returned. A tuple (possible only as a
        keyword argument) must have length equal to the number of outputs.
    where : array_like, optional
        Values of True indicate to calculate the ufunc at that position, values
        of False indicate to leave the value in the output alone.
    **kwargs

    Returns
    -------
    angle : Tensor
        The angle of the ray intersecting the unit circle at the given
        `x`-coordinate in radians [0, pi]. If `x` is a scalar then a
        scalar is returned, otherwise an array of the same shape as `x`
        is returned.

    See Also
    --------
    cos, arctan, arcsin

    Notes
    -----
    `arccos` is a multivalued function: for each `x` there are infinitely
    many numbers `z` such that `cos(z) = x`. The convention is to return
    the angle `z` whose real part lies in `[0, pi]`.

    For real-valued input data types, `arccos` always returns real output.
    For each value that cannot be expressed as a real number or infinity,
    it yields ``nan`` and sets the `invalid` floating point error flag.

    For complex-valued input, `arccos` is a complex analytic function that
    has branch cuts `[-inf, -1]` and `[1, inf]` and is continuous from
    above on the former and from below on the latter.

    The inverse `cos` is also known as `acos` or cos^-1.

    References
    ----------
    M. Abramowitz and I.A. Stegun, "Handbook of Mathematical Functions",
    10th printing, 1964, pp. 79. http://www.math.sfu.ca/~cbm/aands/

    Examples
    --------
    We expect the arccos of 1 to be 0, and of -1 to be pi:
    >>> import mars.tensor as mt

    >>> mt.arccos([1, -1]).execute()
    array([ 0.        ,  3.14159265])

    Plot arccos:

    >>> import matplotlib.pyplot as plt
    >>> x = mt.linspace(-1, 1, num=100)
    >>> plt.plot(x.execute(), mt.arccos(x).execute())
    >>> plt.axis('tight')
    >>> plt.show()
    """
    op = TensorArccos(**kwargs)
    return op(x, out=out, where=where)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List

from .... import oscar as mo
from ....lib.aio import alru_cache
from ..supervisor.tracker import LifecycleTrackerActor
from .core import AbstractLifecycleAPI


class LifecycleAPI(AbstractLifecycleAPI):
    def __init__(
        self,
        session_id: str,
        lifecycle_tracker_ref: mo.ActorRefType[LifecycleTrackerActor],
    ):
        self._session_id = session_id
        self._lifecycle_tracker_ref = lifecycle_tracker_ref

    @classmethod
    @alru_cache(cache_exceptions=False)
    async def create(cls, session_id: str, address: str) -> "LifecycleAPI":
        """
        Create Lifecycle API.

        Parameters
        ----------
        session_id : str
            Session ID.
        address : str
            Supervisor address.

        Returns
        -------
        lifecycle_api
            Lifecycle API.
        """
        lifecycle_tracker_ref = await mo.actor_ref(
            address, LifecycleTrackerActor.gen_uid(session_id)
        )
        return LifecycleAPI(session_id, lifecycle_tracker_ref)

    @mo.extensible
    async def track(self, tileable_key: str, chunk_keys: List[str]):
        """
        Track tileable.

        Parameters
        ----------
        tileable_key : str
            Tileable key.
        chunk_keys : list
            List of chunk keys.
        """
        return await self._lifecycle_tracker_ref.track(tileable_key, chunk_keys)

    @track.batch
    async def batch_track(self, args_list, kwargs_list):
        tracks = []
        for args, kwargs in zip(args_list, kwargs_list):
            tracks.append(self._lifecycle_tracker_ref.track.delay(*args, **kwargs))
        return await self._lifecycle_tracker_ref.track.batch(*tracks)

    async def incref_tileables(
        self, tileable_keys: List[str], counts: List[int] = None
    ):
        """
        Incref tileables.

        Parameters
        ----------
        tileable_keys : list
             List of tileable keys.
        counts: list
            List of ref count.
        """
        return await self._lifecycle_tracker_ref.incref_tileables(
            tileable_keys, counts=counts
        )

    async def decref_tileables(
        self, tileable_keys: List[str], counts: List[int] = None
    ):
        """
        Decref tileables.

        Parameters
        ----------
        tileable_keys : list
            List of tileable keys.
        counts: list
            List of ref count.
        """
        return await self._lifecycle_tracker_ref.decref_tileables(tileable_keys)

    async def get_tileable_ref_counts(self, tileable_keys: List[str]) -> List[int]:
        """
        Get ref counts of tileables.

        Parameters
        ----------
        tileable_keys : list
            List of tileable keys.

        Returns
        -------
        ref_counts : list
            List of ref counts.
        """
        return await self._lifecycle_tracker_ref.get_tileable_ref_counts(tileable_keys)

    async def incref_chunks(self, chunk_keys: List[str], counts: List[int] = None):
        """
        Incref chunks.

        Parameters
        ----------
        chunk_keys : list
            List of chunk keys.
        counts: list
            List of ref count.
        """
        return await self._lifecycle_tracker_ref.incref_chunks(
            chunk_keys, counts=counts
        )

    async def decref_chunks(self, chunk_keys: List[str], counts: List[int] = None):
        """
        Decref chunks

        Parameters
        ----------
        chunk_keys : list
            List of chunk keys.
        counts: list
            List of ref count.
        """
        return await self._lifecycle_tracker_ref.decref_chunks(
            chunk_keys, counts=counts
        )

    async def get_chunk_ref_counts(self, chunk_keys: List[str]) -> List[int]:
        """
        Get ref counts of chunks.

        Parameters
        ----------
        chunk_keys : list
            List of chunk keys.

        Returns
        -------
        ref_counts : list
            List of ref counts.
        """
        return await self._lifecycle_tracker_ref.get_chunk_ref_counts(chunk_keys)

    async def get_all_chunk_ref_counts(self) -> Dict[str, int]:
        """
        Get all chunk keys' ref counts.

        Returns
        -------
        key_to_ref_counts: dict
        """
        return await self._lifecycle_tracker_ref.get_all_chunk_ref_counts()


class MockLifecycleAPI(LifecycleAPI):
    @classmethod
    async def create(cls, session_id: str, address: str) -> "LifecycleAPI":
        from ..supervisor.service import LifecycleSupervisorService

        service = LifecycleSupervisorService({}, address)
        await service.create_session(session_id)
        return await super().create(session_id=session_id, address=address)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools

import numpy as np

from ... import opcodes as OperandDef
from ...serialization.serializables import KeyField, Int32Field
from ...utils import has_unknown_shape
from ..core import TENSOR_TYPE
from ...lib.sparse import diag as sparse_diag
from ...lib.sparse.core import issparse, get_array_module, get_sparse_module
from ...lib import sparse
from ..core import TensorOrder
from ..array_utils import create_array
from .core import TensorHasInput
from .zeros import TensorZeros
from .array import tensor


def _get_diag_shape(v_shape, k):
    size_0, size_1 = 0, 0
    if k > 0:
        size_1 += k
    elif k < 0:
        size_0 -= k
    size = min(v_shape[0] - size_0, v_shape[1] - size_1)
    return (size,)


class TensorDiagBase:
    __slots__ = ()

    def to_chunk_op(self, *args):
        op = self.copy().reset_key()
        (k,) = args
        op._k = k
        return op

    @classmethod
    def _get_nsplits(cls, op):
        raise NotImplementedError

    @classmethod
    def _get_chunk(cls, op, chunk_k, chunk_shape, chunk_idx):
        raise NotImplementedError

    @classmethod
    def tile(cls, op):
        if op.inputs:
            if has_unknown_shape(*op.inputs):
                yield
        tensor = op.outputs[0]

        # op can be TensorDiag or TensorEye
        k = op.k
        nsplits = op._get_nsplits(op)

        fx = lambda x, y: x - y + k
        cum_size = [np.cumsum(s).tolist() for s in nsplits]
        out_chunks = []
        for out_idx in itertools.product(*[range(len(s)) for s in nsplits]):
            i, j = out_idx
            ld_pos = cum_size[0][i] - 1, cum_size[1][j] - nsplits[1][j]
            ru_pos = cum_size[0][i] - nsplits[0][i], cum_size[1][j] - 1

            ld_fx = fx(*ld_pos)
            ru_fx = fx(*ru_pos)

            chunk_shape = (nsplits[0][i], nsplits[1][j])
            if (ld_fx > 0 and ru_fx > 0) or (ld_fx < 0 and ru_fx < 0):
                # does not cross, fill with zeros
                chunk_op = TensorZeros(
                    dtype=op.dtype,
                    gpu=op.gpu,
                    sparse=op.sparse,
                    shape=chunk_shape,
                    order=tensor.order.value,
                )
                chunk = chunk_op.new_chunk(None, shape=chunk_shape, index=out_idx)
            else:
                lu_pos = ru_pos[0], ld_pos[1]
                chunk_k = fx(*lu_pos)
                chunk = op._get_chunk(op, chunk_k, chunk_shape, out_idx)

            out_chunks.append(chunk)

        new_op = op.copy()
        return new_op.new_tensors(
            op.inputs, tensor.shape, chunks=out_chunks, nsplits=nsplits
        )


class TensorDiag(TensorDiagBase, TensorHasInput):
    _op_type_ = OperandDef.TENSOR_DIAG

    _input = KeyField("input")
    _k = Int32Field("k")

    def __init__(self, k=None, **kw):
        super().__init__(_k=k, **kw)

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        if self.dtype is None:
            self._dtype = self.input.dtype

    def to_chunk_op(self, *args):
        return TensorDiagBase.to_chunk_op(self, *args)

    @classmethod
    def _get_nsplits(cls, op):
        assert op.input.ndim == 1
        k = op.k
        nsplits_1d = op.input.nsplits[0]
        nsplit_0, nsplit_1 = list(nsplits_1d), list(nsplits_1d)
        if k > 0:
            nsplit_0.append(k)
            nsplit_1.insert(0, k)
        elif k < 0:
            nsplit_0.insert(0, abs(k))
            nsplit_1.append(abs(k))
        return nsplit_0, nsplit_1

    @classmethod
    def _get_chunk(cls, op, chunk_k, chunk_shape, chunk_idx):
        assert chunk_shape[0] == chunk_shape[1]
        input_idx = chunk_idx[1] if op.k < 0 else chunk_idx[0]
        input_chunk = op.inputs[0].cix[
            input_idx,
        ]
        op = TensorDiag(k=chunk_k, dtype=op.dtype, gpu=op.gpu, sparse=op.sparse)
        return op.new_chunk([input_chunk], shape=chunk_shape, index=chunk_idx)

    def __call__(self, v, shape, chunk_size=None):
        return self.new_tensor(
            [v], shape, raw_chunk_size=chunk_size, order=TensorOrder.C_ORDER
        )

    @classmethod
    def tile(cls, op):
        tensor = op.outputs[0]

        v = op.input
        k = op.k
        idx = itertools.count(0)
        if v.ndim == 2:
            if has_unknown_shape(*op.inputs):
                yield
            chunks = []
            nsplit = []

            fx = lambda x, y: x - y + k
            in_nsplits = v.nsplits
            cum_size = [np.cumsum(s).tolist() for s in in_nsplits]
            for c in v.chunks:
                i, j = c.index
                ld_pos = cum_size[0][i] - 1, cum_size[1][j] - in_nsplits[1][j]
                ru_pos = cum_size[0][i] - in_nsplits[0][i], cum_size[1][j] - 1

                ld_fx = fx(*ld_pos)
                ru_fx = fx(*ru_pos)

                if (ld_fx > 0 and ru_fx > 0) or (ld_fx < 0 and ru_fx < 0):
                    continue

                lu_pos = ru_pos[0], ld_pos[1]
                chunk_k = fx(*lu_pos)

                chunk_shape = _get_diag_shape(c.shape, chunk_k)
                chunk_idx = (next(idx),)
                chunk_op = op.to_chunk_op(chunk_k)
                chunk = chunk_op.new_chunk(
                    [c], shape=chunk_shape, index=chunk_idx, order=tensor.order
                )
                nsplit.append(chunk_shape[0])
                chunks.append(chunk)

            new_op = op.copy()
            return new_op.new_tensors(
                op.inputs,
                op.outputs[0].shape,
                order=tensor.order,
                chunks=chunks,
                nsplits=(tuple(nsplit),),
            )
        else:
            return (yield from super().tile(op))

    @property
    def k(self):
        return getattr(self, "_k", 0)

    @classmethod
    def execute(cls, ctx, op):
        chunk = op.outputs[0]
        if op.sparse:
            ctx[chunk.key] = sparse.diag(ctx[op.inputs[0].key], k=op.k, gpu=op.gpu)
        else:
            ctx[chunk.key] = create_array(op)("diag", ctx[op.inputs[0].key], k=op.k)


def diag(v, k=0, sparse=None, gpu=None, chunk_size=None):
    """
    Extract a diagonal or construct a diagonal tensor.

    See the more detailed documentation for ``mt.diagonal`` if you use this
    function to extract a diagonal and wish to write to the resulting tensor

    Parameters
    ----------
    v : array_like
        If `v` is a 2-D tensor, return its `k`-th diagonal.
        If `v` is a 1-D tensor, return a 2-D tensor with `v` on the `k`-th
        diagonal.
    k : int, optional
        Diagonal in question. The default is 0. Use `k>0` for diagonals
        above the main diagonal, and `k<0` for diagonals below the main
        diagonal.
    sparse: bool, optional
        Create sparse tensor if True, False as default
    gpu : bool, optional
        Allocate the tensor on GPU if True, False as default
    chunk_size : int or tuple of int or tuple of ints, optional
        Desired chunk size on each dimension

    Returns
    -------
    out : Tensor
        The extracted diagonal or constructed diagonal tensor.

    See Also
    --------
    diagonal : Return specified diagonals.
    diagflat : Create a 2-D array with the flattened input as a diagonal.
    trace : Sum along diagonals.
    triu : Upper triangle of a tensor.
    tril : Lower triangle of a tensor.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> x = mt.arange(9).reshape((3,3))
    >>> x.execute()
    array([[0, 1, 2],
           [3, 4, 5],
           [6, 7, 8]])

    >>> mt.diag(x).execute()
    array([0, 4, 8])
    >>> mt.diag(x, k=1).execute()
    array([1, 5])
    >>> mt.diag(x, k=-1).execute()
    array([3, 7])

    >>> mt.diag(mt.diag(x)).execute()
    array([[0, 0, 0],
           [0, 4, 0],
           [0, 0, 8]])

    """
    if not isinstance(v, TENSOR_TYPE):
        tensor_v = tensor(v)
        if tensor_v.issparse():
            xps = get_sparse_module(tensor_v.data)
            v = xps.csr_matrix(
                (tensor_v.op.data, tensor_v.op.indices, tensor_v.op.indptr),
                tensor_v.shape,
            )
            diag_v = sparse_diag(v, k=k)
        else:
            v = tensor(v).op.data
            diag_v = get_array_module(v).diag(v, k=k)
        sparse = sparse if sparse is not None else issparse(v)
        return tensor(diag_v, gpu=gpu, sparse=sparse, chunk_size=chunk_size)

    sparse = sparse if sparse is not None else v.issparse()

    if v.ndim == 1:
        shape = (v.size + abs(k),) * 2
    elif v.ndim == 2:
        shape = _get_diag_shape(v.shape, k)
    else:
        raise ValueError("Input must be 1- or 2-d.")

    op = TensorDiag(k, dtype=v.dtype, gpu=gpu, sparse=sparse)
    return op(v, shape)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from abc import ABC, abstractmethod
from typing import List, Dict

from ..typing import BandType, SessionType
from ..storage.base import StorageLevel


class Context(ABC):
    """
    Context that providing API that can be
    used inside `tile` and `execute`.
    """

    prev = None
    current = None

    def __init__(
        self,
        session_id: str = None,
        supervisor_address: str = None,
        worker_address: str = None,
        local_address: str = None,
        band: BandType = None,
    ):
        if session_id is None:
            # try to get session id from environment
            session_id = os.environ.get("MARS_SESSION_ID")
            if session_id is None:
                raise ValueError("session_id should be provided to create a context")
        if supervisor_address is None:
            # try to get supervisor address from environment
            supervisor_address = os.environ.get("MARS_SUPERVISOR_ADDRESS")
            if supervisor_address is None:
                raise ValueError(
                    "supervisor_address should be provided to create a context"
                )

        self.session_id = session_id
        self.supervisor_address = supervisor_address
        self.worker_address = worker_address
        self.local_address = local_address
        self.band = band

    @abstractmethod
    def get_current_session(self) -> SessionType:
        """
        Get current session

        Returns
        -------
        session
        """

    @abstractmethod
    def get_local_host_ip(self) -> str:
        """
        Get local worker's host ip

        Returns
        -------
        host_ip : str
        """

    @abstractmethod
    def get_supervisor_addresses(self) -> List[str]:
        """
        Get supervisor addresses.

        Returns
        -------
        supervisor_addresses : list
        """

    @abstractmethod
    def get_worker_addresses(self) -> List[str]:
        """
        Get worker addresses.

        Returns
        -------
        worker_addresses : list
        """

    @abstractmethod
    def get_total_n_cpu(self) -> int:
        """
        Get number of cpus.

        Returns
        -------
        number_of_cpu: int
        """

    @abstractmethod
    def get_slots(self) -> int:
        """
        Get num of slots of current band

        Returns
        -------
        number_of_bands: int
        """

    @abstractmethod
    def get_chunks_result(self, data_keys: List[str], fetch_only: bool = False) -> List:
        """
        Get result of chunks.

        Parameters
        ----------
        data_keys : list
            Data keys.
        fetch_only : bool
            If fetch_only, only fetch data but not return.

        Returns
        -------
        results : list
            Result of chunks if not fetch_only, else return None
        """

    @abstractmethod
    def get_chunks_meta(
        self, data_keys: List[str], fields: List[str] = None, error="raise"
    ) -> List[Dict]:
        """
        Get meta of chunks.

        Parameters
        ----------
        data_keys : list
            Data keys.
        fields : list
            Fields to filter.
        error : str
            raise, ignore

        Returns
        -------
        meta_list : list
            Meta list.
        """

    @abstractmethod
    def get_storage_info(self, address: str, level: StorageLevel):
        """
        Get the customized storage backend info of requested storage backend level at given worker.

        Parameters
        ----------
        address: str
            The worker address.
        level: StorageLevel
            The storage level to fetch the backend info.

        Returns
        -------
        info: dict
            Customized storage backend info dict of all workers. The key is
            worker address, the value is the backend info dict.
        """

    @abstractmethod
    def create_remote_object(self, name: str, object_cls, *args, **kwargs):
        """
        Create remote object.

        Parameters
        ----------
        name : str
            Object name.
        object_cls
            Object class.
        args
        kwargs

        Returns
        -------
        ref
        """

    @abstractmethod
    def get_remote_object(self, name: str):
        """
        Get remote object

        Parameters
        ----------
        name : str
            Object name.

        Returns
        -------
        ref
        """

    @abstractmethod
    def destroy_remote_object(self, name: str):
        """
        Destroy remote object.

        Parameters
        ----------
        name : str
            Object name.
        """

    @abstractmethod
    def register_custom_log_path(
        self,
        session_id: str,
        tileable_op_key: str,
        chunk_op_key: str,
        worker_address: str,
        log_path: str,
    ):
        """
        Register custom log path.

        Parameters
        ----------
        session_id : str
            Session ID.
        tileable_op_key : str
            Key of tileable's op.
        chunk_op_key : str
            Kye of chunk's op.
        worker_address : str
            Worker address.
        log_path : str
            Log path.
        """

    def new_custom_log_dir(self) -> str:
        """
        New custom log dir.

        Returns
        -------
        custom_log_dir : str
            Custom log dir.
        """

    def set_running_operand_key(self, session_id: str, op_key: str):
        """
        Set key of running operand.

        Parameters
        ----------
        session_id : str
        op_key : str
        """

    def set_progress(self, progress: float):
        """
        Set progress of running operand.

        Parameters
        ----------
        progress : float
        """

    def __enter__(self):
        Context.prev = Context.current
        Context.current = self

    def __exit__(self, *_):
        Context.current = Context.prev
        Context.prev = None


def set_context(context: Context):
    Context.current = context


def get_context() -> Context:
    return Context.current



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pickle  # nosec  # pylint: disable=import_pickle

import numpy as np

from .... import opcodes
from ....core import OutputType, recursive_tile
from ....dataframe.core import DATAFRAME_TYPE, SERIES_TYPE
from ....serialization.serializables import BytesField, DictField, TupleField
from ...operands import LearnOperand, LearnOperandMixin


class StatsModelsPredict(LearnOperand, LearnOperandMixin):
    _op_code_ = opcodes.STATSMODELS_PREDICT

    _model_results = BytesField(
        "model_results", on_serialize=pickle.dumps, on_deserialize=pickle.loads
    )
    _predict_args = TupleField("predict_args")
    _predict_kwargs = DictField("predict_kwargs")

    def __init__(
        self, model_results=None, predict_args=None, predict_kwargs=None, **kw
    ):
        super().__init__(
            _model_results=model_results,
            _predict_args=predict_args,
            _predict_kwargs=predict_kwargs,
            **kw
        )

    @property
    def model_results(self):
        return self._model_results

    @property
    def predict_args(self) -> tuple:
        return self._predict_args

    @property
    def predict_kwargs(self) -> dict:
        return self._predict_kwargs

    def __call__(self, exog):
        if isinstance(exog, (DATAFRAME_TYPE, SERIES_TYPE)):
            self._output_types = [OutputType.series]
            kwargs = dict(
                shape=exog.shape[:1],
                index_value=exog.index_value,
                dtype=np.dtype("float"),
            )
        else:
            self._output_types = [OutputType.tensor]
            kwargs = dict(
                shape=exog.shape[:1],
                dtype=np.dtype("float"),
            )
        return self.new_tileable([exog], **kwargs)

    @classmethod
    def tile(cls, op: "StatsModelsPredict"):
        exog = op.inputs[0]
        out = op.outputs[0]

        if exog.ndim > 1 and exog.chunk_shape[1] > 1:
            exog = yield from recursive_tile(exog.rechunk({1: exog.shape[1]}))

        chunks = []
        for in_chunk in exog.chunks:
            if isinstance(exog, (DATAFRAME_TYPE, SERIES_TYPE)):
                kwargs = dict(
                    index=in_chunk.index[:1],
                    shape=in_chunk.shape[:1],
                    index_value=in_chunk.index_value,
                    dtype=out.dtype,
                )
            else:
                kwargs = dict(
                    index=in_chunk.index[:1],
                    shape=in_chunk.shape[:1],
                    dtype=out.dtype,
                )

            new_op = op.copy().reset_key()
            chunks.append(new_op.new_chunk([in_chunk], **kwargs))

        new_op = op.copy().reset_key()
        return new_op.new_tileables(
            [exog], chunks=chunks, nsplits=(exog.nsplits[0],), **out.params
        )

    @classmethod
    def execute(cls, ctx, op: "StatsModelsPredict"):
        in_data = ctx[op.inputs[0].key]
        ctx[op.outputs[0].key] = op.model_results.model.predict(
            in_data, *op.predict_args, **op.predict_kwargs
        )



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# noinspection PyUnresolvedReferences
from ..core import GroupBy, DataFrameGroupBy, SeriesGroupBy


def _install():
    from ..core import DATAFRAME_TYPE, SERIES_TYPE, GROUPBY_TYPE, DATAFRAME_GROUPBY_TYPE
    from .core import groupby
    from .aggregation import agg
    from .apply import groupby_apply
    from .cum import cumcount, cummin, cummax, cumprod, cumsum
    from .getitem import df_groupby_getitem
    from .head import head
    from .transform import groupby_transform
    from .sample import groupby_sample
    from .fill import ffill, bfill, fillna

    for cls in DATAFRAME_TYPE:
        setattr(cls, "groupby", groupby)

    for cls in SERIES_TYPE:
        setattr(cls, "groupby", groupby)

    for cls in GROUPBY_TYPE:
        setattr(cls, "agg", agg)
        setattr(cls, "aggregate", agg)

        setattr(cls, "sum", lambda groupby, **kw: agg(groupby, "sum", **kw))
        setattr(cls, "prod", lambda groupby, **kw: agg(groupby, "prod", **kw))
        setattr(cls, "max", lambda groupby, **kw: agg(groupby, "max", **kw))
        setattr(cls, "min", lambda groupby, **kw: agg(groupby, "min", **kw))
        setattr(cls, "count", lambda groupby, **kw: agg(groupby, "count", **kw))
        setattr(cls, "size", lambda groupby, **kw: agg(groupby, "size", **kw))
        setattr(cls, "mean", lambda groupby, **kw: agg(groupby, "mean", **kw))
        setattr(cls, "var", lambda groupby, **kw: agg(groupby, "var", **kw))
        setattr(cls, "std", lambda groupby, **kw: agg(groupby, "std", **kw))
        setattr(cls, "all", lambda groupby, **kw: agg(groupby, "all", **kw))
        setattr(cls, "any", lambda groupby, **kw: agg(groupby, "any", **kw))
        setattr(cls, "skew", lambda groupby, **kw: agg(groupby, "skew", **kw))
        setattr(cls, "kurt", lambda groupby, **kw: agg(groupby, "kurt", **kw))
        setattr(cls, "kurtosis", lambda groupby, **kw: agg(groupby, "kurtosis", **kw))
        setattr(cls, "sem", lambda groupby, **kw: agg(groupby, "sem", **kw))
        setattr(cls, "nunique", lambda groupby, **kw: agg(groupby, "nunique", **kw))

        setattr(cls, "apply", groupby_apply)
        setattr(cls, "transform", groupby_transform)

        setattr(cls, "cumcount", cumcount)
        setattr(cls, "cummin", cummin)
        setattr(cls, "cummax", cummax)
        setattr(cls, "cumprod", cumprod)
        setattr(cls, "cumsum", cumsum)

        setattr(cls, "head", head)

        setattr(cls, "sample", groupby_sample)

        setattr(cls, "ffill", ffill)
        setattr(cls, "bfill", bfill)
        setattr(cls, "backfill", bfill)
        setattr(cls, "fillna", fillna)

    for cls in DATAFRAME_GROUPBY_TYPE:
        setattr(cls, "__getitem__", df_groupby_getitem)


_install()
del _install



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

cdef class Resource:
    cdef readonly:
        float num_cpus
        float num_gpus
        float mem_bytes

    def __init__(self, float num_cpus=0, float num_gpus=0, float mem_bytes=0):
        self.num_cpus = num_cpus
        self.num_gpus = num_gpus
        self.mem_bytes = mem_bytes

    def __eq__(self, Resource other):
        cdef bint ret = (
            self.mem_bytes == other.mem_bytes
            and self.num_gpus == other.num_gpus
            and self.num_cpus == other.num_cpus
        )
        return ret

    cdef bint _le(self, Resource other) nogil:
        # memory first, then gpu, cpu last
        cdef bint ret = (
            self.mem_bytes <= other.mem_bytes
            and self.num_gpus <= other.num_gpus
            and self.num_cpus <= other.num_cpus
        )
        return ret

    def __gt__(self, Resource other):
        return not self._le(other)

    def __le__(self, Resource other):
        return self._le(other)

    def __add__(self, Resource other):
        return Resource(
            num_cpus=self.num_cpus + other.num_cpus,
            num_gpus=self.num_gpus + other.num_gpus,
            mem_bytes=self.mem_bytes + other.mem_bytes,
        )

    def __sub__(self, Resource other):
        return Resource(
            num_cpus=self.num_cpus - other.num_cpus,
            num_gpus=self.num_gpus - other.num_gpus,
            mem_bytes=self.mem_bytes - other.mem_bytes,
        )

    def __neg__(self):
        return Resource(
            num_cpus=-self.num_cpus,
            num_gpus=-self.num_gpus,
            mem_bytes=-self.mem_bytes,
        )

    def __repr__(self):
        return f"Resource(num_cpus={self.num_cpus}, num_gpus={self.num_gpus}, mem_bytes={self.mem_bytes})"

ZeroResource = Resource(num_cpus=0, num_gpus=0, mem_bytes=0)
